/* timing defines ----------------------------------------------------------- */

/*
    The following section does two things :
        (1) it defines the clock resolution and the period of the RT task
        (2) it maps the abstract time values of the automaton to real values
        
    The clock resolution is defined as an integral number of *clocks ticks*.
    On x86 systems, there are 1193180 *clock ticks* in one second, wich is 
    because the Intel 8254 PIT has a base frequency of 1193180 Hz.
    
    The clock resolution is defined in code with the BEAT define.
    For example, #define BEAT 3000 will make the PIT interrupt every 3000 ticks,
    thus updating every clock every (3000/1193180) seconds,
    which is approximately 2,514289546 milliseconds.
    With such a setting, DELTA_P could safely be considered with a value of
    3 milliseconds for example, and actually any value *safely larger* than
    2,514289546 milliseconds would fit.
    
    An important remark about the BEAT define : *do not* set it too low !
    #define BEAT 1 would make the PIT interrupt *every tick* resulting in 
    a complete machine freeze. The safe range depends largely on what the CPU
    can handle; my personal experience is that any setting lower than 10
    should be considered dangerous (#define BEAT 10 makes for a beat of
    approximately 8,380965152 microseconds).
    
    The period of the generated real-time task is defined in code with
    the PERIOD define, and is expressed in an integral number of *beats*.
    For example, #define PERIOD 100 along with #define BEAT 3000 will make
    the period of the task equal to 100*(3000/1193180) seconds, which is
    approximately 251,428954558 milliseconds. With such a setting, DELTA_T
    could safely be considered 300 milliseconds (WARNING : this only holds
    for uniprocessor systems).
    
    The WIDENER define is used for guard enlargement. It must be defined as
    an integral number of beats. For example : a #define WIDENER 10 makes
    a clock "equal to" 250 milliseconds iff its value is anywhere in the
    interval [224.857104544, 275.142895456] (assuming #define BEAT 3000).
    
    The TIME_UNIT define is used to map the abstract time values of the 
    automaton to real values, and is expressed as an integral number of beats.
    For example, with #define TIME_UNIT 20 and #define BEAT 3000, the guard
    "x = 2" will be true when the value x is in the interval
    [40 beats - WIDENER, 40 beats + WIDENER] with 40 beats being
    approximately 100,57158184 milliseconds.
*/

// Requirements : DELTA < U/4 && DELTA > DELTA_T + 2 DELTA_D + 4 DELTA_P
// DELTA_P = 10 µs
// DELTA_T = 10 ms
// DELTA_D = 1 ms
// => DELTA > 10ms + 2ms + 40µs = 12,04 ms
// -> DELTA = 15 ms
// => 15 ms < U / 4
//    U > 60 ms

#define TIME_UNIT   7000    // 64,5 ms -> U > 60 ms OK
#define WIDENER     1001    // DELTA_S = DELTA_T (1000) + DELTA_P (1)
#define PERIOD      1000    // 9,2 ms -> DELTA_T = 10 ms OK
#define BEAT        11      // 9,2 µs -> DELTA_P = 10 µs OK

// Effective bitrate = 1 bit every 4*64,5 ms => 3,87 bps

/* misc parameters ---------------------------------------------------------- */

#define STACK_SIZE 10000
#define USE_FPU 1

/* global decoration -------------------------------------------------------- */
#define TRUE 1 
#define FALSE 0

#define BASE_ADDRESS	0x378
#define DATA_REGISTER	(BASE_ADDRESS)
#define	STAT_REGISTER	(BASE_ADDRESS + 1)
#define CTRL_REGISTER	(BASE_ADDRESS + 2)

int var_p, var_doublezero;
int data[] = {
	1, 1, 1, 1, 1, 1, 1, 1,
	1, 0, 1, 0, 1, 0, 1, 1,
	1, 0, 1, 0, 1, 1, 1, 1,
	1, 0, 1, 0, 1, 0, 1, 1,
	1, 0, 1, 1, 1, 0, 1, 1,
	1, 0, 1, 0, 1, 0, 1, 1,
	1, 1, 1, 0, 1, 0, 1, 1,	
	1, 1, 1, 1, 1, 1, 1, 1,
	0, 0
};

const int data_length = 8*8+2;
int data_index;

int sequence_finished()
{
   return data_index >= data_length;
}

/* spectre globals----------------------------------------------------------- */

RTIME clock_x;

int AASAP_equals(RTIME a, RTIME b);
int AASAP_biggerthan(RTIME a, RTIME other);
int AASAP_smallerthan(RTIME a, RTIME other);

void sender_task_main(int data);

void (*current_location)(void);
RTIME current_time;
int running;

/* readings headers --------------------------------------------------------- */

int sender_read_i();
int sender_read_p();
int sender_read_doublezero();

/* writings headers --------------------------------------------------------- */

void sender_write_p(int value);
void sender_write_doublezero(int value);

/* orders headers ----------------------------------------------------------- */

void sender_order_up(void);
void sender_order_down(void);

/* locations headers -------------------------------------------------------- */

void sender_location_idle(void);
void sender_location_oneSent(void);
void sender_location_waitingOne(void);
void sender_location_zeroSent(void);
void sender_location_waitingZero(void);

/* transitions headers ------------------------------------------------------ */

void sender_transition_initially(void);
void sender_transition_idle_to_oneSent_1(void);
void sender_transition_oneSent_to_waitingOne_1(void);
void sender_transition_oneSent_to_zeroSent_2(void);
void sender_transition_oneSent_to_idle_3(void);
void sender_transition_waitingOne_to_oneSent_1(void);
void sender_transition_zeroSent_to_waitingZero_1(void);
void sender_transition_zeroSent_to_oneSent_2(void);
void sender_transition_zeroSent_to_idle_3(void);
void sender_transition_zeroSent_to_idle_4(void);
void sender_transition_waitingZero_to_zeroSent_1(void);

/* guards headers ----------------------------------------------------------- */

void sender_guard_idle_to_oneSent_1(void);
void sender_guard_oneSent_to_waitingOne_1(void);
void sender_guard_oneSent_to_zeroSent_2(void);
void sender_guard_oneSent_to_idle_3(void);
void sender_guard_waitingOne_to_oneSent_1(void);
void sender_guard_zeroSent_to_waitingZero_1(void);
void sender_guard_zeroSent_to_oneSent_2(void);
void sender_guard_zeroSent_to_idle_3(void);
void sender_guard_zeroSent_to_idle_4(void);
void sender_guard_waitingZero_to_zeroSent_1(void);

/* checks headers ----------------------------------------------------------- */

// TODO

/* decorations headers ------------------------------------------------------ */

// TODO

/* restriction headers ------------------------------------------------------ */

// TODO
