module Main where

import SyntaxTree
import List

genComment :: String -> String
genComment s = "/* " ++ s ++ " " ++ (replicate n '-') ++ " */"
    where n = 73 - (length s)

getDecorationNameStr :: Spectre -> String
getDecorationNameStr s = let (Identifier d _ _) = getDecorationName s in d

readingFunctionName :: Spectre -> String -> String
readingFunctionName s var = "spectre_read_" ++ var

writingFunctionName :: Spectre -> String -> String
writingFunctionName s var = "spectre_write_" ++ var

genExpression :: Spectre -> Expression -> String
genExpression s (Addition e t) = ge ++ "+" ++ gt
    where ge = genExpression s e
          gt = genTerm s t

genExpression s (Substraction e t) = ge ++ "-" ++ gt
    where ge = genExpression s e
          gt = genTerm s t

genExpression s (Term t) = genTerm s t

genTerm :: Spectre -> Term -> String
genTerm s (Multiplication t f) = gt ++ "*" ++ gf
    where gt = genTerm s t
          gf = genFactor s f
          
genTerm s (Division t f) = gt ++ "/" ++ gf
    where gt = genTerm s t
          gf = genFactor s f
          
genTerm s (Factor f) = genFactor s f

genFactor :: Spectre -> Factor -> String
genFactor s (Number i) = show i
genFactor s (Symbol (Identifier i _ _)) = (readingFunctionName s i) ++ "()"
genFactor s (Parenthesis exp) = "(" ++ (genExpression s exp) ++ ")"

genClockExpression :: Spectre -> Expression -> String
genClockExpression s (Addition e t) = ge ++ "+" ++ gt
    where ge = genClockExpression s e
          gt = genClockTerm s t

genClockExpression s (Substraction e t) = ge ++ "-" ++ gt
    where ge = genClockExpression s e
          gt = genClockTerm s t

genClockExpression s (Term t) = genClockTerm s t

genClockTerm :: Spectre -> Term -> String
genClockTerm s (Multiplication t f) = gt ++ "*" ++ gf
    where gt = genClockTerm s t
          gf = genClockFactor s f
          
genClockTerm s (Division t f) = gt ++ "/" ++ gf
    where gt = genClockTerm s t
          gf = genClockFactor s f
          
genClockTerm s (Factor f) = genClockFactor s f

genClockFactor :: Spectre -> Factor -> String
genClockFactor s (Number i) = "( TIME_UNIT * "++ show i ++ ")"
genClockFactor s (Symbol (Identifier i _ _)) = (readingFunctionName s i) ++ "()"
genClockFactor s (Parenthesis exp) = "(" ++ (genClockExpression s exp) ++ ")"

genTransitionId :: Spectre -> (String, Transition) -> String
genTransitionId s a = doGen $ project a
    where project :: (String, Transition) -> (String, String, Int)
          project (a, t@(Transition _ _ _ (Identifier b _ _))) = (a,b, tNum t a)
          tNum :: Transition -> String -> Int
          tNum t a = getTransitionNumber s t a
          doGen :: (String, String, Int) -> String
          doGen (a,b,i) = a ++ "_to_" ++ b ++ "_" ++ (show i)

areLocsConnected :: Spectre -> String -> String -> Bool
areLocsConnected s a b = let ts = getAllLocTransitions s in doTest ts
    where doTest :: [(String, Transition)] -> Bool
          doTest [] = False
          doTest (x:xs) = if doTest' x then True else doTest xs
          doTest' :: (String, Transition) -> Bool
          doTest'(c, (Transition _ _ _ (Identifier d _ _))) = a == c && b == d

combine :: [a] -> [(a,a)]
combine l = combine' l
    where combine' [] = []
          combine' (x:xs) = (doCombine x l) ++ (combine' xs)
          doCombine _ [] = []
          doCombine str (x:xs) = [(str,x)] ++ (doCombine str xs)

buildAllLocsPairs :: Spectre -> [(String,String)]
buildAllLocsPairs s = let ls = getAllLocations s in doBuild ls
    where doBuild :: [Location] -> [(String,String)]
          doBuild ls = combine $ map project ls
          project :: Location -> String
          project (Location (Identifier n _ _) _) = n

getAllConnectedLocs :: Spectre -> [(String,String)]
getAllConnectedLocs s = let ls = buildAllLocsPairs s in doGet ls
    where doGet :: [(String, String)] -> [(String,String)]
          doGet ls = filter f ls
          f :: (String, String) -> Bool
          f (a,b) = areLocsConnected s a b

findReadingCodeOf :: Spectre -> String -> Maybe Code
findReadingCodeOf s x = let ds = getAllDecorationItems s in doFind ds
    where doFind :: [DecorationItem] -> Maybe Code
          doFind [] = Nothing
          doFind ((Reading (Identifier i _ _) c):xs) = if i == x then (Just c)
                                                       else doFind xs
          doFind (_:xs) = doFind xs

findWritingCodeOf :: Spectre -> String -> Maybe Code
findWritingCodeOf s x = let ds = getAllDecorationItems s in doFind ds
    where doFind :: [DecorationItem] -> Maybe Code
          doFind [] = Nothing
          doFind ((Writing (Identifier i _ _) _ c):xs) = if i == x then (Just c)
                                                       else doFind xs
          doFind (_:xs) = doFind xs

genParameterOf :: Spectre -> String -> String
genParameterOf s x = let ds = getAllDecorationItems s in doFind ds
    where doFind :: [DecorationItem] -> String
          doFind [] = "foo"
          doFind ((Writing i p c):xs) = if n i == x then n p else doFind xs
          doFind (_:xs) = doFind xs
          n :: Identifier -> String
          n (Identifier i _ _) = i
          
findLocToLocDecorationCode :: Spectre -> String -> String -> Int -> Maybe Code
findLocToLocDecorationCode s a b i = let d = getAllDecorationItems s in doFind d
    where n :: Identifier -> String
          n (Identifier i _ _) = i
          doFind :: [DecorationItem] -> Maybe Code
          doFind [] = Nothing
          doFind ((FromLocToLoc c d c1 c2):xs) = if ok then res else doFind xs
              where ok = n c == a && n d == b
                    res = if i == 1 then (Just c1) else (Just c2)
          doFind (_:xs) = doFind xs                   
                    
findAnyToLocDecorationCode :: Spectre -> String -> Int -> Maybe Code
findAnyToLocDecorationCode s a i = let d = getAllDecorationItems s in doFind d
    where n :: Identifier -> String
          n (Identifier i _ _) = i
          doFind :: [DecorationItem] -> Maybe Code
          doFind [] = Nothing
          doFind ((FromAnyToLoc b c1 c2):xs) = if ok then res else doFind xs
              where ok = n b == a
                    res = if i == 1 then (Just c1) else (Just c2)
          doFind (_:xs) = doFind xs
                    
findLocToAnyDecorationCode :: Spectre -> String -> Int -> Maybe Code
findLocToAnyDecorationCode s a i = let d = getAllDecorationItems s in doFind d
    where n :: Identifier -> String
          n (Identifier i _ _) = i
          doFind :: [DecorationItem] -> Maybe Code
          doFind [] = Nothing
          doFind ((FromLocToAny b c1 c2):xs) = if ok then res else doFind xs
              where ok = n b == a
                    res = if i == 1 then (Just c1) else (Just c2)
          doFind (_:xs) = doFind xs                    
                    
findAnyToAnyDecorationCode :: Spectre -> Int -> Maybe Code
findAnyToAnyDecorationCode s i = let d = getAllDecorationItems s in doFind d
    where n :: Identifier -> String
          n (Identifier i _ _) = i
          doFind :: [DecorationItem] -> Maybe Code
          doFind [] = Nothing
          doFind ((FromAnyToAny c1 c2):xs) = if i == 1 then (Just c1) 
                                                       else (Just c2)
          doFind (_:xs) = doFind xs                                                       

isClock :: Spectre -> String -> Bool
isClock s i = let dc = findDeclaredClocks s in doTest $ project dc
    where project :: Maybe [Identifier] -> [String]
          project Nothing = []
          project (Just []) = []
          project (Just ((Identifier j _ _):xs)) = [j] ++ (project (Just xs))
          doTest :: [String] -> Bool
          doTest [] = False
          doTest (x:xs) = if x == i then True else doTest xs
          
genAssignments :: Spectre -> [Assignment] -> String
genAssignments s a = unlines $ map doGen a
    where doGen :: Assignment -> String
          doGen (Assignment (Identifier i _ _) exp) = if isClock s i
                                                      then genClockAss i exp
                                                      else genVarAss i exp
          genClockAss :: String -> Expression -> String
          genClockAss x e = (writingFunctionName s x) ++ "(" ++ 
                            (genClockExpression s e) ++ ");"
          genVarAss :: String -> Expression -> String
          genVarAss x e = (writingFunctionName s x) ++ "(" ++ 
                          (genExpression s e) ++ ");"                   
--------------------------------------------------------------------------------

genRtaiIncludes :: String
genRtaiIncludes = unlines [
    genComment "RTAI includes",
    "#include <linux/module.h>",
    "#include <rtai.h>",
    "#include <rtai_sched.h>",
    "#include <rtai_sem.h>",
    "#include <rtai_fifos.h>"]

genUserDefines :: String
genUserDefines = unlines [
    genComment "user defines",
    "#define TIME_UNIT   <define this>",
    "#define WIDENER     <define this>",
    "#define PERIOD      <define this>",
    "#define BEAT        <define this>"]
    
genSystemParameters :: String
genSystemParameters = unlines [
    genComment "system defines",
    "#define STACK_SIZE 10000",
    "#define USE_FPU 1",
    "#define TASK_PRIORITY 128 /* 0 = highest, 255 = lowest */"]

genUserGlobals :: Spectre -> String
genUserGlobals s = let (Global c) = getGlobalDecoration s
    in if c == Nop then none else code c
    where none = genComment "nop for global decoration"
          code (Code s) = unlines [genComment "global decoration", s]

genClocksDeclaration :: Spectre -> String
genClocksDeclaration s = let d = findClocksDeclaration s in doGen d
    where doGen :: Maybe Declaration -> String
          doGen Nothing = ""
          doGen (Just (ClocksDeclaration ids)) = unlines $ doGen' ids
          doGen' :: [Identifier] -> [String]
          doGen' [] = []
          doGen' ((Identifier x _ _):xs) = [declaration x] ++ (doGen' xs)
          declaration :: String -> String
          declaration x = "RTIME spectre_clock_" ++ x ++ ";"                
          
genSpectreGlobals :: Spectre -> String
genSpectreGlobals s = unlines [
    genComment "spectre gobals",
    genClocksDeclaration s,
    "int spectre_equals(RTIME a, RTIME b);",
    "int spectre_largerthan(RTIME a, RTIME other);",
    "int spectre_smallerthan(RTIME a, RTIME other);",
    "RT_TASK spectre_task;",
    "void spectre_task_main(int data);",
    "void (*spectre_current_location)(void);",
    "RTIME spectre_current_time;",
    "int spectre_running;",
    "void spectre_check_for_events(void);",
    pendings]
    where pendings = let de = findDeclaredEvents s in 
                     unlines $ map doGen $ project de
          project :: Maybe [Identifier] -> [String]
          project Nothing = []
          project (Just ((Identifier i _ _):xs)) = [i] ++ (project $ Just xs)
          project (Just []) = []
          doGen :: String -> String
          doGen x = "int spectre_" ++ x ++ "_is_pending = 0;"
    
genReadingsHeaders :: Spectre -> String
genReadingsHeaders s = unlines [
    genComment "variables readings headers",
    genVarReadingsHeaders s,
    genComment "clocks readings headers",
    genClockReadingsHeaders s]
    
genVarReadingsHeaders :: Spectre -> String
genVarReadingsHeaders s = let vs = findDeclaredVariables s in doGen vs
    where doGen :: Maybe [Identifier] -> String
          doGen Nothing = "// no declared variables //"
          doGen (Just vars) = unlines $ doGen' vars
          doGen' :: [Identifier] -> [String]
          doGen' [] = []
          doGen' ((Identifier x _ _):xs) = [gen x] ++ (doGen' xs)
          gen :: String -> String
          gen x = "int spectre_read_" ++ x ++ "(void);"
        
genClockReadingsHeaders :: Spectre -> String    
genClockReadingsHeaders s = let vs = findDeclaredClocks s in doGen vs
    where doGen :: Maybe [Identifier] -> String
          doGen Nothing = "// no declared variables //"
          doGen (Just vars) = unlines $ doGen' vars
          doGen' :: [Identifier] -> [String]
          doGen' [] = []
          doGen' ((Identifier x _ _):xs) = [gen x] ++ (doGen' xs)
          gen :: String -> String
          gen x = "RTIME spectre_read_" ++ x ++ "(void);"

genWritingsHeaders :: Spectre -> String
genWritingsHeaders s = unlines [
    genComment "variables writings headers",
    genVarWritingsHeaders s,
    genComment "clocks writings headers",
    genClockWritingsHeaders s]

genVarWritingsHeaders :: Spectre -> String
genVarWritingsHeaders s = let vs = findDeclaredVariables s in doGen vs
    where doGen :: Maybe [Identifier] -> String
          doGen Nothing = "// no declared variables //"
          doGen (Just vars) = unlines $ doGen' vars
          doGen' :: [Identifier] -> [String]
          doGen' [] = []
          doGen' ((Identifier x _ _):xs) = [gen x] ++ (doGen' xs)
          gen :: String -> String
          gen x = "void spectre_write_" ++ x ++ "(int value);"
        
genClockWritingsHeaders :: Spectre -> String    
genClockWritingsHeaders s = let vs = findDeclaredClocks s in doGen vs
    where doGen :: Maybe [Identifier] -> String
          doGen Nothing = "// no declared variables //"
          doGen (Just vars) = unlines $ doGen' vars
          doGen' :: [Identifier] -> [String]
          doGen' [] = []
          doGen' ((Identifier x _ _):xs) = [gen x] ++ (doGen' xs)
          gen :: String -> String
          gen x = "void spectre_write_" ++ x ++ "(RTIME value);"

genLocationsHeaders :: Spectre -> String
genLocationsHeaders s = let ls = getAllDeclaredLocations s in doGen $ names ls
    where doGen :: [String] -> String
          doGen a = unlines $ [genComment "location headers"] ++ (doGen' a)
          doGen' :: [String] -> [String]
          doGen' [] = []
          doGen' (x:xs) = [gen x] ++ (doGen' xs)
          gen :: String -> String
          gen x = "void spectre_location_" ++ x ++ "(void);"
          name :: Identifier -> String
          name (Identifier n _ _) = n
          names :: [Identifier] -> [String]
          names = map name
          
genTransitionsHeaders :: Spectre -> String
genTransitionsHeaders s = let ts = getAllLocTransitions s
    in unlines $ [genComment "transitions headers", init] ++ (map doGen ts)
    where init :: String
          init = "void spectre_transition_initially(void);"
          doGen :: (String, Transition) -> String
          doGen (a,t) = "void spectre_transition_" ++ id ++ "(void);"
              where id = genTransitionId s (a,t)

genUpdatesHeaders :: Spectre -> String
genUpdatesHeaders s = let ts = getAllLocTransitions s
    in unlines $ [genComment "updates headers", init] ++ (map doGen ts)
    where init :: String
          init = "void spectre_update_initially(void);"
          doGen :: (String, Transition) -> String
          doGen (a,t) = "void spectre_update_" ++ id ++ "(void);"
              where id = genTransitionId s (a,t)

genGuardsHeaders :: Spectre -> String
genGuardsHeaders s = let ts = getAllLocTransitions s
    in unlines $ [genComment "guards headers"] ++ (map doGen ts)
    where doGen :: (String, Transition) -> String
          doGen (a,t) = "int spectre_guard_" ++ id ++ "(void);"
              where id = genTransitionId s (a,t)

genChecksHeaders :: Spectre -> String
genChecksHeaders s = let ts = getAllLocTransitions s
    in unlines $ [genComment "checks headers"] ++ (map doGen ts)
    where doGen :: (String, Transition) -> String
          doGen (a,t) = "int spectre_check_" ++ id ++ "(void);"
              where id = genTransitionId s (a,t)

genLocToLocDecorationsHeaders :: Spectre -> String
genLocToLocDecorationsHeaders s = let cl = getAllConnectedLocs s 
    in unlines $ (map doGen1 cl) ++ [""] ++ (map doGen2 cl)
    where doGen1 :: (String,String) -> String
          doGen1 (a,b) = "void spectre_d1_" ++ a ++ "_to_" ++ b ++ "(void);"
          doGen2 :: (String,String) -> String
          doGen2 (a,b) = "void spectre_d2_" ++ a ++ "_to_" ++ b ++ "(void);"
          
genAnyToLocDecorationsHeaders :: Spectre -> String
genAnyToLocDecorationsHeaders s = let ls = getAllLocations s
    in unlines $ (map doGen1 $ map project ls) ++ [""] ++ 
                 (map doGen2 $ map project ls)
    where project :: Location -> String
          project (Location (Identifier l _ _) _) = l
          doGen1 :: String -> String
          doGen1 l = "void spectre_d1_any_to_" ++ l ++ "(void);"
          doGen2 :: String -> String
          doGen2 l = "void spectre_d2_any_to_" ++ l ++ "(void);"
          
genLocToAnyDecorationsHeaders :: Spectre -> String
genLocToAnyDecorationsHeaders s = let ls = getAllLocations s
    in unlines $ (map doGen1 $ map project ls) ++ [""] ++ 
                 (map doGen2 $ map project ls)
    where project :: Location -> String
          project (Location (Identifier l _ _) _) = l
          doGen1 :: String -> String
          doGen1 l = "void spectre_d1_" ++ l ++ "_to_any(void);"
          doGen2 :: String -> String
          doGen2 l = "void spectre_d2_" ++ l ++ "_to_any(void);"          
                    
genAnyToAnyDecorationsHeaders :: String
genAnyToAnyDecorationsHeaders = unlines [
    "void spectre_d1_any_to_any(void);",
    "void spectre_d2_any_to_any(void);"]    

genDecorationsHeaders :: Spectre -> String
genDecorationsHeaders s = unlines [
    genComment "decorations headers",
    genLocToLocDecorationsHeaders s,
    genAnyToLocDecorationsHeaders s,
    genLocToAnyDecorationsHeaders s,
    genAnyToAnyDecorationsHeaders]

genLocToLocRestrictionsHeaders :: Spectre -> String
genLocToLocRestrictionsHeaders s = let cl = getAllConnectedLocs s 
    in unlines $ map doGen cl
    where doGen :: (String,String) -> String
          doGen (a,b) = "int spectre_r_" ++ a ++ "_to_" ++ b ++ "(void);"

genAnyToLocRestrictionsHeaders :: Spectre -> String
genAnyToLocRestrictionsHeaders s = let ls = getAllLocations s
    in unlines $ map doGen $ map project ls               
    where project :: Location -> String
          project (Location (Identifier l _ _) _) = l
          doGen :: String -> String
          doGen l = "int spectre_r_any_to_" ++ l ++ "(void);"
          
genLocToAnyRestrictionsHeaders :: Spectre -> String
genLocToAnyRestrictionsHeaders s = let ls = getAllLocations s
    in unlines $ map doGen $ map project ls               
    where project :: Location -> String
          project (Location (Identifier l _ _) _) = l
          doGen :: String -> String
          doGen l = "int spectre_r_"++ l ++ "_to_any(void);"          
          
genAnyToAnyRestrictionsHeaders :: String
genAnyToAnyRestrictionsHeaders = "int spectre_r_any_to_any(void);"

genRestrictionsHeaders :: Spectre -> String
genRestrictionsHeaders s = unlines [
    genComment "restrictions headers",
    genLocToLocRestrictionsHeaders s,
    genAnyToLocRestrictionsHeaders s,
    genLocToAnyRestrictionsHeaders s,
    genAnyToAnyRestrictionsHeaders]

genOrdersHeaders :: Spectre -> String
genOrdersHeaders s = let os = findDeclaredOrders s
    in unlines $ [genComment "orders headers"] ++ 
                 (map (doGen 1) $ project os) ++ (map (doGen 2) $ project os)
    where project :: Maybe [Identifier] -> [String]
          project Nothing = []
          project (Just ((Identifier i _ _):xs)) = [i] ++ (project $ Just xs)
          project (Just []) = []
          doGen :: Int -> String -> String
          doGen i x = "void spectre_d" ++ (show i) ++ "_" ++ x ++ "(void);"

genInternalsHeaders :: Spectre -> String
genInternalsHeaders s = let os = findDeclaredInternals s
    in unlines $ [genComment "internals headers"] ++
                 (map (doGen 1) $ project os) ++ (map (doGen 2) $ project os)
    where project :: Maybe [Identifier] -> [String]
          project Nothing = []
          project (Just ((Identifier i _ _):xs)) = [i] ++ (project $ Just xs)
          project (Just []) = []
          doGen :: Int -> String -> String
          doGen i x = "void spectre_d" ++ (show i) ++ "_" ++ x ++ "(void);"
          
genEventsHeaders :: Spectre -> String
genEventsHeaders s = let os = findDeclaredEvents s
    in unlines $ [genComment "events headers"] ++
                 (map doGenR $ project os) ++  (map (doGen 1) $ project os) ++ 
                 (map (doGen 2) $ project os)
    where project :: Maybe [Identifier] -> [String]
          project Nothing = []
          project (Just ((Identifier i _ _):xs)) = [i] ++ (project $ Just xs)
          project (Just []) = []
          doGenR :: String -> String
          doGenR x = "int spectre_poll_" ++ x ++ "(void);"
          doGen :: Int -> String -> String
          doGen i x = "void spectre_d" ++ (show i) ++ "_" ++ x ++ "(void);"
         
--------------------------------------------------------------------------------

genReadingsImps :: Spectre -> String
genReadingsImps s =  unlines [
    genComment "variables readings implementation",
    genVarReadingsImps s,
    genComment "clocks readings implementation",
    genClocksReadingsImps s]
    
genVarReadingsImps :: Spectre -> String
genVarReadingsImps s = let vs = findDeclaredVariables s in doGen vs
    where doGen :: Maybe [Identifier] -> String
          doGen Nothing = "// no declared variables //"
          doGen (Just vars) = unlines $ map doGen' $ map project vars
          doGen' :: String -> String
          doGen' x = "int spectre_read_" ++ x ++ "() {" ++ (genCode x) ++ "}"
          genCode :: String -> String
          genCode x = let rc = findReadingCodeOf s x in genCode' rc
          genCode' :: Maybe Code -> String
          genCode' (Nothing) = "return 0;"
          genCode' (Just Nop) = "return 0;"
          genCode' (Just (Code c)) = c
          project :: Identifier -> String
          project (Identifier i _ _) = i
                  
genClocksReadingsImps :: Spectre -> String
genClocksReadingsImps s = let vs = findDeclaredClocks s in doGen vs
    where doGen :: Maybe [Identifier] -> String
          doGen Nothing = "// no declared clocks //"
          doGen (Just clocks) = unlines $ map doGen' $ map project clocks
          doGen' :: String -> String
          doGen' x = "RTIME spectre_read_" ++ x ++ "(void) {" ++ (genCode x) ++ "}"
          genCode :: String -> String
          genCode x = "return spectre_current_time - spectre_clock_" ++ x ++ ";"
          project :: Identifier -> String
          project (Identifier i _ _) = i

genWritingsImps :: Spectre -> String
genWritingsImps s =  unlines [
    genComment "variables writings implementation",
    genVarWritingsImps s,
    genComment "clocks writings implementation",
    genClocksWritingsImps s]
    
genVarWritingsImps :: Spectre -> String
genVarWritingsImps s = let vs = findDeclaredVariables s in doGen vs
    where doGen :: Maybe [Identifier] -> String
          doGen Nothing = "// no declared variables //"
          doGen (Just vars) = unlines $ map doGen' $ map project vars
          doGen' :: String -> String
          doGen' x = "void spectre_write_" ++ x ++"(int " ++
                     (genParameterOf s x) ++ ") {" ++ (genCode x) ++ "}"                     
          genCode :: String -> String
          genCode x = let rc = findWritingCodeOf s x in genCode' rc
          genCode' :: Maybe Code -> String
          genCode' (Nothing) = ";"
          genCode' (Just Nop) = ";"
          genCode' (Just (Code c)) = c
          project :: Identifier -> String
          project (Identifier i _ _) = i
                  
genClocksWritingsImps :: Spectre -> String
genClocksWritingsImps s = let vs = findDeclaredClocks s in doGen vs
    where doGen :: Maybe [Identifier] -> String
          doGen Nothing = "// no declared clocks //"
          doGen (Just clocks) = unlines $ map doGen' $ map project clocks
          doGen' :: String -> String
          doGen' x = "void spectre_write_" ++ x ++ "(RTIME newtime) {" ++
                     (genCode x) ++ "}"
          genCode :: String -> String
          genCode x = "spectre_clock_" ++ x ++ 
                      " = spectre_current_time - newtime;"
          project :: Identifier -> String
          project (Identifier i _ _) = i

genLocToLocDecorationsImps :: Spectre -> String
genLocToLocDecorationsImps s = let cl = getAllConnectedLocs s 
    in unlines $ (map (doGen 1) cl) ++ [""] ++ (map (doGen 2) cl)
    where doGen :: Int -> (String,String) -> String
          doGen i (a,b) = "void spectre_d" ++ (show i) ++ "_" ++ a ++ "_to_" ++
                           b ++ "() {" ++ (genCode a b i) ++ "}"
          genCode :: String -> String -> Int -> String
          genCode a b i = let c = findLocToLocDecorationCode s a b i in proj c
          proj :: Maybe Code -> String
          proj (Nothing) = ";"
          proj (Just Nop) = ";"
          proj (Just (Code c)) = c
          
genAnyToLocDecorationsImps :: Spectre -> String
genAnyToLocDecorationsImps s = let ls = getAllLocations s
    in unlines $ (map (doGen 1) $ map project ls) ++ [""] ++ 
                 (map (doGen 2) $ map project ls)
    where project :: Location -> String
          project (Location (Identifier l _ _) _) = l
          doGen :: Int -> String -> String
          doGen i l = "void spectre_d" ++ (show i) ++ "_any_to_" ++ l ++ 
                      "(void) {" ++ (genCode i l) ++ "}"
          genCode :: Int -> String -> String
          genCode i l = let c = findAnyToLocDecorationCode s l i in proj c
          proj :: Maybe Code -> String
          proj (Nothing) = ";"
          proj (Just Nop) = ";"
          proj (Just (Code c)) = c        
          
genLocToAnyDecorationsImps :: Spectre -> String
genLocToAnyDecorationsImps s = let ls = getAllLocations s
    in unlines $ (map (doGen 1) $ map project ls) ++ [""] ++ 
                 (map (doGen 2) $ map project ls)
    where project :: Location -> String
          project (Location (Identifier l _ _) _) = l
          doGen :: Int -> String -> String
          doGen i l = "void spectre_d" ++ (show i) ++ "_" ++ l ++ 
                      "_to_any(void) {" ++ (genCode i l) ++ "}"
          genCode :: Int -> String -> String
          genCode i l = let c = findLocToAnyDecorationCode s l i in proj c
          proj :: Maybe Code -> String
          proj (Nothing) = ";"
          proj (Just Nop) = ";"
          proj (Just (Code c)) = c          
                    
genAnyToAnyDecorationsImps :: Spectre -> String
genAnyToAnyDecorationsImps s = unlines [doGen 1, doGen 2]
    where doGen :: Int -> String
          doGen i = "void spectre_d" ++ (show i) ++ "_any_to_any(void) { " ++
                    (genCode i) ++ "}"
          genCode :: Int -> String
          genCode i = let c = findAnyToAnyDecorationCode s i in proj c
          proj :: Maybe Code -> String
          proj (Nothing) = ";"
          proj (Just Nop) = ";"
          proj (Just (Code c)) = c 

genDecorationsImps :: Spectre -> String
genDecorationsImps s = unlines [
    genComment "decorations implementation",
    genLocToLocDecorationsImps s,
    genAnyToLocDecorationsImps s,
    genLocToAnyDecorationsImps s,
    genAnyToAnyDecorationsImps s]

genGuardsImps :: Spectre -> String
genGuardsImps s = let ts = getAllLocTransitions s
    in unlines $ [genComment "guards implementation"] ++ (map doGen ts)
    where proj :: Transition -> [Predicate]
          proj (Transition (Predicates ps) _ _ _) = ps
          doGen :: (String, Transition) -> String
          doGen (a,t) = "int spectre_guard_" ++ id ++ 
                        "(void) {" ++ header ++ middle ++ footer ++ "}"
                        where id = genTransitionId s (a,t)
                              header = unlines [
                                       "RTIME spectre_t1, spectre_t2;",
                                       "int spectre_e1, spectre_e2;",
                                       "int spectre_res = 1, spectre_tmp;"]
                              middle = concat $ map (genPredicate s) $ proj t
                              footer = "return spectre_res;"

genPredicate :: Spectre -> Predicate -> String
genPredicate s p = if isClockPredicate s p then genClockPredicate s p
                                           else genVarPredicate s p

genClockPredicate :: Spectre -> Predicate -> String
genClockPredicate s (Predicate e1 comp e2) = unlines [ce1, ce2, ccomp, footer]
    where ce1 = "spectre_t1 = " ++ (genClockExpression s e1) ++ ";"
          ce2 = "spectre_t2 = " ++ (genClockExpression s e2) ++ ";"
          ccomp = "spectre_tmp = " ++ (genComp comp) ++ 
                  "(spectre_t1, spectre_t2);"
          footer = "spectre_res = spectre_tmp && spectre_res;"
          genComp :: Comparator -> String
          genComp EqualTo = "spectre_equals"
          genComp SmallerThan = "spectre_smallerthan"
          genComp LargerThan = "spectre_largerthan"
          
genVarPredicate :: Spectre -> Predicate -> String
genVarPredicate s (Predicate e1 comp e2) = unlines [ce1, ce2, ccomp, footer]
    where ce1 = "spectre_e1 = " ++ (genExpression s e1) ++ ";"
          ce2 = "spectre_e2 = " ++ (genExpression s e2) ++ ";"
          ccomp = "spectre_tmp = spectre_e1 " ++ (genComp comp) ++
                  " spectre_e2;"
          footer = "spectre_res = spectre_tmp && spectre_res;"
          genComp :: Comparator -> String
          genComp EqualTo = "=="
          genComp SmallerThan = "<="
          genComp LargerThan = ">="
                                           
genLocToLocRestrictionsImps :: Spectre -> String
genLocToLocRestrictionsImps s = let cl = getAllConnectedLocs s 
    in unlines $ map doGen cl
    where doGen :: (String,String) -> String
          doGen (a,b) = "int spectre_r_" ++ a ++ "_to_" ++ b ++ "(void) {" ++
                        (genCode a b) ++ "}"
          genCode :: String -> String -> String                        
          genCode a b = let ds = getAllDecorationItems s in doGenCode a b ds
          n :: Identifier -> String
          n (Identifier i _ _) = i
          proj :: Code -> String
          proj Nop = "return 1;"
          proj (Code c) = c
          doGenCode :: String -> String -> [DecorationItem] -> String
          doGenCode a b [] = "return 1;"
          doGenCode a b ((Restriction (RestrictFromLocToLoc i j) c):xs) = res
              where res = if n i == a && n j == b 
                          then proj c 
                          else doGenCode a b xs
          doGenCode a b (_:xs) = doGenCode a b xs

genAnyToLocRestrictionsImps :: Spectre -> String
genAnyToLocRestrictionsImps s = let ls = getAllLocations s
    in unlines $ map doGen $ map project ls               
    where project :: Location -> String
          project (Location (Identifier l _ _) _) = l
          doGen :: String -> String
          doGen l = "int spectre_r_any_to_" ++ l ++ "(void) {" ++
                    (genCode l) ++ "}"
          genCode :: String -> String
          genCode l = let ds = getAllDecorationItems s in doGenCode l ds
          n :: Identifier -> String
          n (Identifier i _ _) = i
          proj :: Code -> String
          proj Nop = "return 1;"
          proj (Code c) = c
          doGenCode :: String -> [DecorationItem] -> String
          doGenCode a [] = "return 1;"
          doGenCode a ((Restriction (RestrictFromAnyToLoc i) c):xs) = res
              where res = if n i == a
                          then proj c 
                          else doGenCode a xs
          doGenCode a (_:xs) = doGenCode a xs
          
          
genLocToAnyRestrictionsImps :: Spectre -> String
genLocToAnyRestrictionsImps s = let ls = getAllLocations s
    in unlines $ map doGen $ map project ls               
    where project :: Location -> String
          project (Location (Identifier l _ _) _) = l
          doGen :: String -> String
          doGen l = "int spectre_r_" ++ l ++ "_to_any(void) {" ++
                    (genCode l) ++ "}"
          genCode :: String -> String
          genCode l = let ds = getAllDecorationItems s in doGenCode l ds
          n :: Identifier -> String
          n (Identifier i _ _) = i
          proj :: Code -> String
          proj Nop = "return 1;"
          proj (Code c) = c
          doGenCode :: String -> [DecorationItem] -> String
          doGenCode a [] = "return 1;"
          doGenCode a ((Restriction (RestrictFromLocToAny i) c):xs) = res
              where res = if n i == a
                          then proj c 
                          else doGenCode a xs
          doGenCode a (_:xs) = doGenCode a xs        
          
genAnyToAnyRestrictionsImps :: Spectre -> String
genAnyToAnyRestrictionsImps s = "int spectre_r_any_to_any(void) {" ++ doGen ++ "}"
    where doGen :: String
          doGen = let ds = getAllDecorationItems s in genCode ds
          proj :: Code -> String
          proj Nop = "return 1;"
          proj (Code c) = c
          genCode :: [DecorationItem] -> String
          genCode [] = "return 1;"
          genCode ((Restriction (RestrictFromAnyToAny) c):xs) = proj c
          genCode (_:xs) = genCode xs
          
genRestrictionsImps :: Spectre -> String
genRestrictionsImps s = unlines [
    genComment "restrictions implementation",
    genLocToLocRestrictionsImps s,
    genAnyToLocRestrictionsImps s,
    genLocToAnyRestrictionsImps s,
    genAnyToAnyRestrictionsImps s]

genUpdatesImps :: Spectre -> String
genUpdatesImps s = let ts = getAllLocTransitions s
    in unlines $ [genComment "updates implementation", init] ++ (map doGen ts)
    where init :: String
          init = "void spectre_update_initially(void) {" ++ 
                 (genAssignments s (getInitialUpdates s)) ++ "}"
          doGen :: (String, Transition) -> String
          doGen (a,t@(Transition _ _ (Assignments ass) _)) = 
                "void spectre_update_" ++ id ++ "(void) {" ++ gen ++ "}"
              where id = genTransitionId s (a,t)
                    gen = genAssignments s ass

genLabelDecoration :: Spectre -> String -> Int -> String
genLabelDecoration s x i = let di = getAllDecorationItems s in doGen di
    where doGen :: [DecorationItem] -> String
          doGen [] = error $ "Decoration for label " ++ x ++ "not found !"
          doGen ((Order (Identifier n _ _) c1 c2):xs) = if n == x 
                                                        then selectCode c1 c2
                                                        else doGen xs
          doGen ((Internal (Identifier n _ _) c1 c2):xs) = if n == x 
                                                           then selectCode c1 c2
                                                           else doGen xs
          doGen ((Event (Identifier n _ _) _ c1 c2):xs) = if n == x 
                                                          then selectCode c1 c2
                                                          else doGen xs
          doGen (_:xs) = doGen xs                                                          
          selectCode :: Code -> Code -> String
          selectCode c1 c2 = if i == 1 then genCode c1 else genCode c2
          genCode :: Code -> String
          genCode Nop = ";"
          genCode (Code s) = s

genOrdersImps :: Spectre -> String
genOrdersImps s = let os = findDeclaredOrders s
    in unlines $ [genComment "orders implementation"] ++ 
                 (map (doGen 1) $ proj os) ++ (map (doGen 2) $ proj os)
    where proj :: Maybe [Identifier] -> [String]
          proj Nothing = []
          proj (Just ((Identifier i _ _):xs)) = [i] ++ (proj $ Just xs)
          proj (Just []) = []
          doGen :: Int -> String -> String
          doGen i x = "void spectre_d" ++ (show i) ++ "_" ++ x ++ "(void) {" ++
                     (genLabelDecoration s x i) ++ "}"
   
genInternalsImps :: Spectre -> String
genInternalsImps s = let os = findDeclaredInternals s
    in unlines $ [genComment "internals implementation"] ++ 
                 (map (doGen 1) $ proj os) ++ (map (doGen 2) $ proj os)
    where proj :: Maybe [Identifier] -> [String]
          proj Nothing = []
          proj (Just ((Identifier i _ _):xs)) = [i] ++ (proj $ Just xs)
          proj (Just []) = []
          doGen :: Int -> String -> String
          doGen i x = "void spectre_d" ++ (show i) ++ "_" ++ x ++ "(void) {" ++
                     (genLabelDecoration s x i) ++ "}"

genEventRestriction :: Spectre -> String -> String
genEventRestriction s x = let di = getAllDecorationItems s in doGen di
    where doGen :: [DecorationItem] -> String
          doGen [] = error $ "Decoration for event " ++ x ++ "not found !"
          doGen ((Event (Identifier n _ _) c _ _):xs) = if n == x 
                                                          then genCode c
                                                          else doGen xs
          doGen (_:xs) = doGen xs                                                          
          genCode :: Code -> String
          genCode Nop = "return 1;"
          genCode (Code s) = s
                     
genEventsImps :: Spectre -> String
genEventsImps s = let os = findDeclaredEvents s
    in unlines $ [genComment "events implementation"] ++ 
                 (map (doGen 1) $ proj os) ++ (map (doGen 2) $ proj os) ++ 
                 (map doGenP $ proj os)
    where proj :: Maybe [Identifier] -> [String]
          proj Nothing = []
          proj (Just ((Identifier i _ _):xs)) = [i] ++ (proj $ Just xs)
          proj (Just []) = []
          doGen :: Int -> String -> String
          doGen i x = "void spectre_d" ++ (show i) ++ "_" ++ x ++ "(void) {" ++
                     (genLabelDecoration s x i) ++ "}"
          doGenP :: String -> String
          doGenP x = "int spectre_poll_" ++ x ++ "(void) {" ++
                     (genEventRestriction s x) ++ "}"
                     
genChecksImps :: Spectre -> String
genChecksImps s = let ts = getAllLocTransitions s
    in unlines $ [genComment "checks implementation"] ++ (map doGen ts)
    where doGen :: (String, Transition) -> String
          doGen (a,t) = "int spectre_check_" ++ id ++ "(void) {" ++ code ++ "}"
              where id = genTransitionId s (a,t)
                    end :: Transition -> String
                    end (Transition _ _ _ (Identifier n _ _)) = n     
                    code = unlines [header, r1, r2, r3, r4, g, footer]
                    header = "int spectre_r1, spectre_r2, spectre_r3, " ++
                             "spectre_r4, spectre_g;"
                    footer = "return (spectre_g && spectre_r1 " ++
                             "&& spectre_r2 && spectre_r3 && spectre_r4" ++
                             evt ++ ");"
                    lab :: Transition -> Label
                    lab (Transition _ l _ _) = l
                    n :: Label -> String
                    n (Label (Identifier l _ _)) = l
                    n _ = ""               
                    evt = if isEvent s (n $ lab t)
                          then "&& (spectre_" ++ (n $ lab t) ++ 
                               "_is_pending == 1)"
                          else ""
                    r1 = "spectre_r1 = spectre_r_" ++ a ++
                         "_to_" ++ (end t) ++ "();"
                    r2 = "spectre_r2 = spectre_r_" ++ a ++ "_to_any();"
                    r3 = "spectre_r3 = spectre_r_any_to_" ++ (end t) ++ "();"
                    r4 = "spectre_r4 = spectre_r_any_to_any();"
                    g = "spectre_g = spectre_guard_" ++ id ++ "();"
                    
isEvent :: Spectre -> String -> Bool
isEvent s x = let de = findDeclaredEvents s in doTest $ project de  
    where project :: Maybe [Identifier] -> [String]
          project Nothing = []
          project (Just []) = []
          project (Just ((Identifier j _ _):xs)) = [j] ++ (project (Just xs))
          doTest :: [String] -> Bool
          doTest [] = False
          doTest (y:ys) = if x == y then True else doTest ys
    
genTransitionsImps :: Spectre -> String
genTransitionsImps s = let ts = getAllLocTransitions s
    in unlines $ [genComment "transitions implementation", init] ++ 
                 (map doGen ts)
    where init :: String
          init = "void spectre_transition_initially(void) { " ++ 
                 "spectre_update_initially(); " ++ 
                 "spectre_current_location = spectre_location_" ++ 
                 (getInitialLocation s) ++ ";}"
          end :: Transition -> String
          end (Transition _ _ _ (Identifier i _ _)) = i
          lab :: Transition -> Label
          lab (Transition _ l _ _) = l
          genDCalls :: Int -> String -> String -> String
          genDCalls i a b = unlines $ map (genDCall i) 
                            [(a,b),(a,"any"), ("any", b), ("any", "any")]
          genDCall :: Int -> (String, String) -> String
          genDCall i (a,b) = "spectre_d" ++ (show i) ++ "_" ++ a ++
                             "_to_" ++ b ++ "();" 
          genLCall :: Int -> Label -> String
          genLCall i (Label (Identifier l _ _)) = "spectre_d" ++ (show i) ++ 
                                                  "_" ++ l ++ "();" 
          genLCall i None = ""
          doGen :: (String, Transition) -> String
          doGen (a,t) = "void spectre_transition_" ++ 
                        (genTransitionId s (a,t)) ++ "(void) { "
                        ++ td1 ++ ld1 ++ u ++ ld2 ++ td2 ++ ass ++ post ++ "}"
               where td1 = genDCalls 1 a (end t)
                     td2 = genDCalls 2 a (end t)
                     ld1 = genLCall 1 (lab t)
                     ld2 = genLCall 2 (lab t)
                     n :: Label -> String
                     n (Label (Identifier l _ _)) = l
                     n _ = ""
                     post = if isEvent s (n $ lab t)
                            then "spectre_" ++ (n $ lab t) ++ "_is_pending = 0;"
                            else ""
                     u = "spectre_update_" ++ (genTransitionId s (a,t)) ++ "();" 
                     ass = "spectre_current_location = " ++ 
                           "spectre_location_" ++ (end t) ++ ";"
                           
genLocationsImps :: Spectre -> String
genLocationsImps s = let ls = getAllDeclaredLocations s in doGen $ names ls
    where doGen :: [String] -> String
          doGen a = unlines $ [genComment "location implementation"] ++ (gen a)
          gen :: [String] -> [String]
          gen [] = []
          gen (x:xs) = [gen' x] ++ (gen xs)
          gen' :: String -> String
          gen' x = "void spectre_location_" ++ x ++ "(void) {" ++ trans x ++ "}"
          trans :: String -> String
          trans x = let lt = getAllLocTransitions s 
                             in unlines $ map t $ f x lt
          f :: String -> [(String, Transition)] -> [(String, Transition)]
          f x ls = filter (\a -> (fst a) == x) ls
          t :: (String, Transition) -> String
          t (a, t) = "if("++ cond ++ "())" ++ tran ++ "();"
              where cond = "spectre_check_" ++ (genTransitionId s (a,t))                     
                    tran = "spectre_transition_" ++ (genTransitionId s (a,t)) 
          name :: Identifier -> String
          name (Identifier n _ _) = n
          names :: [Identifier] -> [String]
          names = map name            
          
genCheckForEvents :: Spectre -> String
genCheckForEvents s =  let de = findDeclaredEvents s in 
                       header ++ (unlines $ map doGen $ project de) ++ footer
    where header = "void spectre_check_for_events(void) {"
          footer = "}"
          project :: Maybe [Identifier] -> [String]
          project Nothing = []
          project (Just ((Identifier i _ _):xs)) = [i] ++ (project $ Just xs)
          project (Just []) = []
          doGen :: String -> String
          doGen x = "if(("++  pending ++ " == 0) && " ++ poll ++ ")" ++
                    pending ++ " = 1;"
              where pending = "spectre_" ++ x ++ "_is_pending"
                    poll = "spectre_poll_" ++ x ++ "()"

genCleanup :: Spectre -> String
genCleanup s = let di = getAllDecorationItems s in doGen di
    where doGen :: [DecorationItem] -> String
          doGen [] = error "Cleanup decoration not found !"
          doGen ((Cleanup c):_) = header ++ (genCode c) ++ footer
          doGen (_:xs) = doGen xs
          header = "void spectre_cleanup(void) {"
          footer = "}"
          genCode :: Code -> String
          genCode Nop = ";"
          genCode (Code c) = c

genStartup :: Spectre -> String
genStartup s = let di = getAllDecorationItems s in doGen di
    where doGen :: [DecorationItem] -> String
          doGen [] = error "Startup decoration not found !"
          doGen ((Startup c):_) = header ++ (genCode c) ++ footer
          doGen (_:xs) = doGen xs
          header = "void spectre_startup(void) {"
          footer = "}"
          genCode :: Code -> String
          genCode Nop = ";"
          genCode (Code c) = c

genSpectreGlobalsImps :: String
genSpectreGlobalsImps = unlines [
    "int spectre_equals(RTIME a, RTIME b)",
    "{",
    "    RTIME lower = b - WIDENER;",
    "    RTIME higher = b + WIDENER;",
    "    return (a > lower && a < higher);",
    "}",
    "",
    "int spectre_largerthan(RTIME a, RTIME other)",
    "{",
    "    other -= WIDENER;",
    "    return (a > other);",
    "}",
    "",
    "int spectre_smallerthan(RTIME a, RTIME other)",
    "{",
    "    other += WIDENER;",
    "    return (a < other);",
    "}"]
          
genEndOfFile :: String
genEndOfFile  = unlines [
    genComment "main function implementation",
    "void spectre_task_main(int data)",
    "{",
    "    spectre_running = 1;",
    "    spectre_startup();",
    "    spectre_transition_initially();",     
    "    while (spectre_running == 1) {",
    "        spectre_current_time = rt_get_time();",
    "        spectre_check_for_events();",
    "        spectre_current_location();",
    "        rt_task_wait_period();",
	"    }",
	"    spectre_cleanup();",
	"}",
	"",
	genComment "kernel module functions",
	"int spectre_init_module(void)",
	"{",
	"    rt_set_periodic_mode();",
    "    start_rt_timer(BEAT);",
	"    rt_task_init (&spectre_task, spectre_task_main, 0, " ++
	"STACK_SIZE, TASK_PRIORITY, USE_FPU, 0);",
	"    rt_task_make_periodic(&spectre_task, rt_get_time() + PERIOD, PERIOD);",
	"    return 0;",
    "}",
    "",
    "void spectre_cleanup_module(void)",
    "{",
	"    stop_rt_timer();",
	"    rt_task_delete(&spectre_task);",
    "}",
    "",
    "module_init(spectre_init_module);",
    "module_exit(spectre_cleanup_module);",
    "MODULE_LICENSE(\"GPL\");",
    ""]

--------------------------------------------------------------------------------

genSpectre :: Spectre -> String
genSpectre s = unlines [
    genRtaiIncludes,
    genUserDefines,                     
    genSystemParameters,              
    genUserGlobals s, 
    genSpectreGlobals s,
    genReadingsHeaders s,              
    genWritingsHeaders s,              
    genOrdersHeaders s,                
    genInternalsHeaders s,              
    genEventsHeaders s,                
    genLocationsHeaders s,              
    genTransitionsHeaders s,           
    genUpdatesHeaders s,                
    genGuardsHeaders s,                 
    genChecksHeaders s,                 
    genDecorationsHeaders s,            
    genRestrictionsHeaders s,          
    genReadingsImps s,                  
    genWritingsImps s,                  
    genDecorationsImps s,               
    genGuardsImps s,                  
    genRestrictionsImps s,             
    genUpdatesImps s,                 
    genOrdersImps s,                   
    genInternalsImps s,               
    genEventsImps s,                 
    genChecksImps s,                 
    genTransitionsImps s,             
    genLocationsImps s,
    genCheckForEvents s,
    genCleanup s,
    genStartup s,
    genSpectreGlobalsImps,
    genEndOfFile]               

getSpectre :: String -> IO Spectre
getSpectre s = return (read s)

main = do str <- getContents
          s <- getSpectre str
          putStr (genSpectre s)
