{
module Main  where
import Tokens
import SyntaxTree
}

%name parser
%tokentype { Token }

%token
    any                 { TokenAny               pos      }
    nop                 { TokenNop               pos      }
    none                { TokenNone              pos      }
    specification       { TokenSpecification     pos      }
    decoration          { TokenDecoration        pos      }
    end                 { TokenEnd               pos      }
    id                  { TokenIdentifier        pos name }
    ','                 { TokenComma             pos      }
    ';'                 { TokenSemiColon         pos      }
    ':'                 { TokenColon             pos      }
    '+'                 { TokenPlus              pos      }
    '-'                 { TokenMinus             pos      }
    '*'                 { TokenTimes             pos      }
    '/'                 { TokenDivide            pos      }
    '('                 { TokenLeftParenthesis   pos      }
    ')'                 { TokenRightParenthesis  pos      }
    '{'                 { TokenLeftCurlyBrace    pos      }
    '}'                 { TokenRightCurlyBrace   pos      }
    ':='                { TokenAssign            pos      }
    '='                 { TokenEquals            pos      }
    '<='                { TokenSmallerThan       pos      }
    '>='                { TokenLargerThan        pos      }
    number              { TokenNumber            pos val  }
    clocks              { TokenClocks            pos      }
    vars                { TokenVars              pos      }
    events              { TokenEvents            pos      }
    internals           { TokenInternals         pos      }
    orders              { TokenOrders            pos      }
    initially           { TokenInitially         pos      }
    location            { TokenLocation          pos      }
    global              { TokenGlobal            pos      }
    startup             { TokenStartup           pos      }
    cleanup             { TokenCleanup           pos      }
    event               { TokenEvent             pos      }
    order               { TokenOrder             pos      }
    internal            { TokenInternal          pos      }
    reading             { TokenReading           pos      }
    writing             { TokenWriting           pos      }
    restrict            { TokenRestrict          pos      }
    code                { TokenCCode             pos code }
    to                  { TokenTo                pos      }
%%

Spectre : Specification Decoration {Spectre $1 $2}

Specification : specification id 
                     Declarations
                     Initially
                     Locations      
                end 
                {Specification (idToId $2) (Declarations $3) $4 (Locations $5)}

Declarations : {-nothing-} {[]}
             | Declaration Declarations {[$1] ++ $2}

Declaration : ClocksDeclaration    {$1}
            | VarsDeclaration      {$1}
            | EventsDeclaration    {$1}
            | OrdersDeclaration    {$1}
            | InternalsDeclaration {$1}

ClocksDeclaration :      clocks    ':' IdList ';'  {ClocksDeclaration $3}
VarsDeclaration :        vars      ':' IdList ';'  {VarsDeclaration $3}
EventsDeclaration :      events    ':' IdList ';'  {EventsDeclaration $3}
OrdersDeclaration :      orders    ':' IdList ';'  {OrdersDeclaration $3}
InternalsDeclaration :   internals ':' IdList ';'  {InternalsDeclaration $3}


IdList : {- nothing -}                        {[]}
       | id IdListTail                        {[(idToId $1)] ++ $2}
                  
IdListTail : {- nothing -}                    {[]}
           | ',' id IdListTail                {[(idToId $2)] ++ $3}

Decoration : decoration id DecorationItems end
           {Decoration (idToId $2) (DecorationItems $3)}

Expression : Expression '+' Term		{ Addition $1 $3 }
           | Expression '-' Term	    { Substraction $1 $3 }
           | Term			            { Term $1 }

Term : Term '*' Factor	{ Multiplication $1 $3 }
     | Term '/' Factor	{ Division $1 $3 }
     | Factor			{ Factor $1 }

Factor : number			        { Number (numberValue $1)}
	 | id			            { Symbol (idToId $1)}
	 | '(' Expression ')'		{ Parenthesis $2 }
	 

Initially : initially id ',' Assignments ';' {Initially (idToId $2) $4} 

Assignments : '{' AssignmentList '}' {Assignments $2}

AssignmentList : {- nothing -}                         {[]}
               | Assignment AssignmentListTail         {[$1] ++ $2}

AssignmentListTail : {- nothing -}                     {[]}
                   | ',' Assignment AssignmentListTail {[$2] ++ $3}
                   
Assignment : id ':=' Expression {Assignment (idToId $1) $3} 

Locations : Location LocationsTail          {[$1] ++ $2}

LocationsTail : {- nothing -}               {[]}
              | Location LocationsTail      {[$1] ++ $2}
              
Location : location id ':' Transitions         
           {Location (idToId $2) (Transitions $4) }

Transitions : {- nothing -}                 {[]}
            | Transition Transitions        {[$1] ++ $2}
            
Transition : Predicates ',' Label ',' Assignments ',' id ';' 
             {Transition $1 $3 $5 (idToId $7)}
             
Label : id   {Label (idToId $1)}             
      | none {None}

Predicates : '{' PredicateList '}'          {Predicates $2}

PredicateList : {- nothing -}               {[]}
              | Predicate PredicateListTail {[$1] ++ $2}
              
PredicateListTail : {- nothing -}                   {[]}
                  | ',' Predicate PredicateListTail {[$2] ++ $3}

Predicate : Expression Comparator Expression        {Predicate $1 $2 $3}                  

Comparator : '='  { EqualTo }
           | '<=' { SmallerThan }
           | '>=' { LargerThan }

DecorationItems : {- nothing -}                    {[]}
                | DecorationItem DecorationItems   {[$1] ++ $2}                

DecorationItem : global Code       
                 {Global $2}
               | startup Code      
                 {Startup $2}
               | cleanup Code      
                 {Cleanup $2}
               | order id Code Code
                 {Order (idToId $2) $3 $4}
               | event id Code Code Code
                 {Event (idToId $2) $3 $4 $5}
               | internal id Code Code
                 {Internal (idToId $2) $3 $4}
               | reading id Code
                 {Reading (idToId $2) $3}
               | writing id '(' id ')' Code
                 {Writing (idToId $2) (idToId $4) $6}
               | id to id Code Code
                 {FromLocToLoc (idToId $1) (idToId $3) $4 $5}
               | any to id Code Code
                 {FromAnyToLoc (idToId $3) $4 $5}
               | id to any Code Code
                 {FromLocToAny (idToId $1) $4 $5}
               | any to any Code Code
                 {FromAnyToAny $4 $5}  
               | restrict RestrictionHeader Code
                 {Restriction $2 $3}

Code : code {Code (codeContents $1)}
     | nop  {Nop}

RestrictionHeader : id to id
                    {RestrictFromLocToLoc (idToId $1) (idToId $3)}
                  | any to id
                    {RestrictFromAnyToLoc (idToId $3)}
                  | id to any
                    {RestrictFromLocToAny (idToId $1)}
                  | any to any
                    {RestrictFromAnyToAny}
                
{
               
idToId :: Token -> Identifier                
idToId (TokenIdentifier (LineCol l c) str) = Identifier str l c
idToId s = error $ "idToId was applied to a non identifier : " ++ (show s)
                
happyError :: [Token] -> a
happyError (x:xs) = syntaxError x

syntaxError :: Token -> a
syntaxError t = syntaxError'(tokenPos(t)) (tokenStr(t))

syntaxError' :: TokenPos -> String -> a
syntaxError' (LineCol l c) str = error msg
 where msg =   "Syntax error at line " ++ (show l) ++
                ", column " ++ (show c) ++
                " : unexpected token [" ++ str ++ "]"
main = do
  s <- getContents
  print (parser(readTokens(s)))

readTokens :: String -> [Token]
readTokens s = (read s)

}
