module Main where

import SyntaxTree

import Control.Monad.State

data Tree a = Node a [Tree a]
            deriving(Eq,Read,Show)

data Tag = Tag    String -- Label
                  String -- Shape
                  
         | NTag   String  -- Label
                  String  -- Shape
                  Integer -- Identifier
         deriving(Eq,Read,Show)

--------------------------------------------------------------------------------

{-
    The following set of functions are used to build a (Tree Tag) structure
    from a Spectre syntax tree.
-}

mkSpectre (Spectre s d) =                                          -- Spectre --
    Node (Tag "Spectre" "ellipse") c
    where c = [mkSpecification s] ++ [mkDecoration d]
    
mkDecoration (Decoration n i) =                                 -- Decoration -- 
    Node (Tag "Decoration" "ellipse") c
    where c = [mkName n] ++ [mkDecorationItems i]
    
mkDecorationItems (DecorationItems i) =                    -- DecorationItems --
    Node (Tag "DecorationItems" "ellipse") c
    where c = map mkDecorationItem i   
    
mkDecorationItem (Global code) =                            -- DecorationItem --
    Node (Tag "Global" "ellipse") [mkCode code]

mkDecorationItem (Startup code) =
    Node (Tag "Startup" "ellipse") [mkCode code]

mkDecorationItem (Cleanup code) =                            
    Node (Tag "Cleanup" "ellipse") [mkCode code]

mkDecorationItem (Order l code) =                            
    Node (Tag "Order" "ellipse") c
    where c = [mkLabel l] ++ [mkCode code]

mkDecorationItem (Event l code) =                            
    Node (Tag "Event" "ellipse") c
    where c = [mkLabel l] ++ [mkCode code]

mkDecorationItem (Internal l code) =                            
    Node (Tag "Internal" "ellipse") c
    where c = [mkLabel l] ++ [mkCode code]
    
mkDecorationItem (Reading v code) =                            
    Node (Tag "Reading" "ellipse") c
    where c = [mkVariable v] ++ [mkCode code]

mkDecorationItem (Writing v p code) =                            
    Node (Tag "Writing" "ellipse") c
    where c = [mkVariable v] ++ [mkParameter p] ++ [mkCode code]

mkDecorationItem (Entering l code) =                            
    Node (Tag "Entering" "ellipse") c
    where c = [mkLocationName l] ++ [mkCode code]
  
mkDecorationItem (Leaving l code) =                            
    Node (Tag "Leaving" "ellipse") c
    where c = [mkLocationName l] ++ [mkCode code]
    
mkDecorationItem (EnteringFrom l1 l2 code) =                            
    Node (Tag "EnteringFrom" "ellipse") c
    where c = [mkLocationName l1] ++ [mkLocationName l2] ++ [mkCode code]    

mkDecorationItem (LeavingTo l1 l2 code) =                            
    Node (Tag "LeavingTo" "ellipse") c
    where c = [mkLocationName l1] ++ [mkLocationName l2] ++ [mkCode code]    

mkDecorationItem (Restriction h code) =                            
    Node (Tag "Restriciton" "ellipse") c
    where c = [mkRestrictionHeader h] ++ [mkCode code]
    
mkRestrictionHeader (RestrictEntering l) =               -- RestrictionHeader --
    Node (Tag "RestrictEntering" "ellipse") [mkLocationName l]

mkRestrictionHeader (RestrictLeaving l) =     
    Node (Tag "RestrictLeaving" "ellipse") [mkLocationName l]
    
mkRestrictionHeader (RestrictEnteringFrom l1 l2) =     
    Node (Tag "RestrictEnteringFrom" "ellipse") c
    where c = [mkLocationName l1] ++ [mkLocationName l2]
    
mkRestrictionHeader (RestrictLeavingTo l1 l2) =     
    Node (Tag "RestrictLeavingTo" "ellipse") c
    where c = [mkLocationName l1] ++ [mkLocationName l2]

mkCode (Code s) =                                                     -- Code --
    Node (Tag "<code>" "box") []
    
mkSpecification (Specification n d i l) =                    -- Specification --
    Node (Tag "Specification" "ellipse") c
    where c = [mkName n] ++ [mkDeclarations d] ++
              [mkInitially i] ++ [mkLocations l]

mkLocations (Locations l) =                                      -- Locations --
    Node (Tag "Locations" "ellipse") c
    where c = map mkLocation l
    
mkLocation (Location n t) =                                       -- Location --
    Node (Tag "Location" "ellipse") c
    where c = [mkLocationName n] ++ [mkTransitions t]

mkTransitions (Transitions t) =                                -- Transitions --
    Node (Tag "Transitions" "ellipse") c
    where c = map mkTransition t
    
mkTransition (Transition p l a n) =                             -- Transition --
    Node (Tag "Transition" "ellipse") c
    where c = [mkPredicates p] ++ [mkLabel l] ++
              [mkAssignments a] ++ [mkLocationName n]

mkPredicates (Predicates p) =                                   -- Predicates --
    Node (Tag "Predicates" "ellipse") c
    where c = map mkPredicate p
    
mkPredicate (Predicate e1 r e2) =                                -- Predicate --
    Node (Tag "Predicate" "ellipse") c
    where c = [mkExpression e1] ++ [mkComparator r] ++ [mkExpression e2]
    
mkComparator (EqualTo) =                                        -- Comparator -- 
    Node (Tag "EqualTo" "ellipse") []
    
mkComparator (SmallerThan) =
    Node (Tag "SmallerThan" "ellipse") []
    
mkComparator (LargerThan) =
    Node (Tag "LargerThan" "ellipse") []
              
mkLabel (Label s) =                                                  -- Label --
    Node (Tag "Label" "ellipse") [mkString s]
    
mkLocationName (LocationName s) =                             -- LocationName --
    Node (Tag "LocationName" "ellipse") [mkString s]

mkInitially (Initially s a) =                                    -- Initially --
    Node (Tag "Initially" "ellipse") c
    where c = (mkString s) : [mkAssignments a]
    
mkAssignments (Assignments a) =                                -- Assignments --
    Node (Tag "Assignments" "ellipse") c
    where c = map mkAssignment a
    
mkAssignment (Assignment s e) =                                 -- Assignment --
    Node (Tag "Assignment" "ellipse") c
    where c = [mkString s] ++ [mkExpression e]
              
mkName (Name s) =                                                     -- Name --
    Node (Tag "Name" "ellipse") [mkString s]
    
mkVariable (Variable s) =                                         -- Variable --
    Node (Tag "Variable" "ellipse") [mkString s]    

mkParameter (Parameter s) =                                      -- Parameter --
    Node (Tag "Parameter" "ellipse") [mkString s]    

mkDeclarations (Declarations d) =                             -- Declarations --
    Node (Tag "Declarations" "ellipse") c
    where c = map mkDeclaration d
    
mkDeclaration (ClocksDeclaration d) =                          -- Declaration --
    Node (Tag "ClocksDeclaration" "ellipse") c
    where c = map mkString d
    
mkDeclaration (OrdersDeclaration d) =
    Node (Tag "OrdersDeclaration" "ellipse") c
    where c = map mkString d
    
mkDeclaration (EventsDeclaration d) =
    Node (Tag "EventsDeclaration" "ellipse") c
    where c = map mkString d

mkDeclaration (InternalsDeclaration d) =
    Node (Tag "InternalsDeclaration" "ellipse") c
        where c = map mkString d

mkDeclaration (VarsDeclaration d) =
    Node (Tag "VarsDeclaration" "ellipse") c
    where c = map mkString d

mkDeclaration (SilentDeclaration d) =
    Node (Tag "SilentDeclaration" "ellipse") c
    where c = [mkString d]

mkString s = Node (Tag s "box") []                                  -- String --

mkFactor (Number i) =                                               -- Factor --
    Node (Tag (show i) "box") []
    
mkFactor (Symbol p) =
    Node (Tag p "box") []
    
mkFactor (Parenthesis e) =                                          
    Node (Tag "Parenthesis" "ellipse") [mkExpression e]

mkTerm (Multiplication t f) =                                         -- Term --
    Node (Tag "Multiplication" "ellipse") c
    where c = [mkTerm t] ++ [mkFactor f]

mkTerm (Division t f) =
    Node (Tag "Division" "ellipse") c
    where c = [mkTerm t] ++ [mkFactor f]

mkTerm (Factor f) =
    Node (Tag "Factor" "ellipse") [mkFactor f]

mkExpression (Addition e t) =                                   -- Expression --
    Node (Tag "Addition" "ellipse") c
    where c = [mkExpression e] ++ [mkTerm t]

mkExpression (Substraction e t) =
    Node (Tag "Substraction" "ellipse") c
    where c = [mkExpression e] ++ [mkTerm t]

mkExpression (Term t) =
    Node (Tag "Term" "ellipse") [mkTerm t]

--------------------------------------------------------------------------------

{-
    The following functions transform a (Tree Tag) tree into a numbered tree.
-}

nTag :: Tag -> Integer -> Tag
nTag (Tag l s) i = NTag l s i
nTag _ _ = error "Cannot number an already numbered tag !"

numberTree t = (evalState (doNumberTree [t]) 0) !! 0

doNumberTree :: [Tree Tag] -> State Integer ([Tree Tag])
doNumberTree [(Node t [])] = do
                           next <- get
                           put (next + 1)
                           return [(Node (nTag t next) [])]

doNumberTree [(Node t (c:cs))] = do
                               nc <- doNumberTree [c]
                               nc2 <- doNumberTree cs
                               next <- get
                               put (next + 1)
                               return [(Node (nTag t next) (nc ++ nc2))]
                               
doNumberTree [] = return []
doNumberTree (x:xs) = do
                    a <- doNumberTree [x]
                    b <- doNumberTree xs
                    return (a ++ b)

--------------------------------------------------------------------------------

{-
    The following functions operate on a numbered tree and are used to
    generate the GraphViz code.
-}

graphViz t = "digraph G { \n" ++ 
             (graphVizHeader t) ++
             "\n" ++
             (graphVizFooter t) ++
             "}"             
            
graphVizHeader (Node t []) =  graphVizHeaderLine t
graphVizHeader (Node t c) = (graphVizHeaderLine t) ++ (graphVizHeaders c)

graphVizHeaders [] = ""
graphVizHeaders (x:xs) = (graphVizHeader x) ++ (graphVizHeaders xs)

graphVizHeaderLine (NTag l s i) =
    "   S" ++ (show i) ++ " [label=\"" ++ l ++ "\", shape=\"" ++ s ++ "\"];\n"


graphVizFooter :: Tree Tag -> String
graphVizFooter (Node t []) = ""
graphVizFooter (Node t c) = (graphVizFooterLines t c) ++ (graphVizFooters c)

graphVizFooters :: [Tree Tag] -> String
graphVizFooters [] = ""
graphVizFooters (x:xs) = (graphVizFooter x) ++ (graphVizFooters xs)

graphVizFooterLines :: Tag -> [Tree Tag] -> String
graphVizFooterLines t [] = ""
graphVizFooterLines t (x:xs) = (graphVizFooterLine t x) ++
                               (graphVizFooterLines t xs)

graphVizFooterLine :: Tag -> Tree Tag -> String
graphVizFooterLine (NTag _ _ i) (Node (NTag _ _ j) _) = 
    "   S" ++ (show i) ++ " -> S" ++ (show j) ++ "\n"

--------------------------------------------------------------------------------
  
readSpectre :: String -> Spectre
readSpectre e = read e
    
main = do
    str <- getContents  
    (putStr.graphViz.numberTree.mkSpectre.readSpectre) str
    putStrLn ""
