module Main where

import System
import SyntaxTree
import Maybe
import List
import IO

{-

 !! FIXMEES !!
 
 (1) assignments involving clocks should *only* involve clocks
 (2?) in predicates, comparing clocks with something else than numbers should
     be forbidden.


            Semantic checker for the Spectre input language
            -----------------------------------------------
         
Here is the list the semantic validation rules needed for code generation.
Rules beginning with an 'S' are related to the specification part, and those
which govern the decoration part start with a 'D'.

--------------------------------------------------------------------------------
            
Rule S01 : There must be 1 and only 1 of each of the following declarations :                     
           clocks, vars, events, internals, orders, silent.

Rule S02 : Every symbol occuring in predicates must be declared as either
           a clock or a variable.

Rule S03 : Every lvalue of an assignment must be declared as either a clock
           or a variable.         
         
Rule S04 : The initial location must be declared as a location.

Rule S05 : In all transitions, the synchronisation label must be declared as
           one of the following : order, silent, internal, event

Rule S06 : In all transitions, the destination location must be declared as 
           a location.
           
Rule S07 : Any identifier can only be defined once and be of one type.
           Types are : clock, variable, event, order, internal, silent, location
           
Rule S08 : Warning if a location has no input transition.

Rule S09 : Predicates should not mix variables and clocks.

--------------------------------------------------------------------------------

Rule D01 : Exactly 1 code fragment must be provided for each one of the
           declared event,internal,order labels.
           
Rule D02 : Every 'reading' and 'writing' items must refer to a *variable* 
           declared in the specification.
           
Rule D03 : The string provided as a parameter to 'writing' cannot be either of 
           the following : variable,clock,location,label.

Rule D04 : Exactly 1 reading must be provided if a variable appears in
           an expression.

Rule D05 : Warning if a reading is provided but the variable never appears in an 
           expression.
        
Rule D06 : Exactly 1 writing must be provided if a variable appears
           as an lvalue.

Rule D07 : Warning if a writing is provided but the variable never appears 
           as an lvalue.

Rule D08 : Transition decorations must refer to declared locations.

Rule D09 : Restrictions decorations must refer to declared locations.
           
Rule D10 : There must be exactly 1 of the following decoration items : 
           global,startup,cleanup
           
Rule D11 : Restrictions and Transition decorations cannot have multiple copies.           
-}

--------------------------------------------------------------------------------

data SemanticCheck = SemanticCheck (Spectre -> Bool) -- checker function
                                   String            -- message
                                   
allSymbolsDeclared :: [String] -> [String] -> Bool
allSymbolsDeclared a [] = True
allSymbolsDeclared a (x:xs) | x `elem` a = allSymbolsDeclared a xs
                            | otherwise  = False

allLabelsDeclared = allSymbolsDeclared
allLocationsDeclared = allSymbolsDeclared
                                                               
--------------------------------------------------------------------------------
-- Rule S01 --
--------------

ruleS01a = SemanticCheck checkS01a msgS01a
ruleS01b = SemanticCheck checkS01b msgS01b
ruleS01c = SemanticCheck checkS01c msgS01c
ruleS01d = SemanticCheck checkS01d msgS01d
ruleS01e = SemanticCheck checkS01e msgS01e

rulesS01 = [ruleS01a, ruleS01b, ruleS01c, ruleS01d, ruleS01e]

checkS01a s = countClocksDeclaration s == 1
checkS01b s = countVarsDeclaration s == 1
checkS01c s = countOrdersDeclaration s == 1
checkS01d s = countInternalsDeclaration s == 1
checkS01e s = countEventsDeclaration s == 1

msgS01a = "[S01a] The clocks declaration"    ++ msgS01footer
msgS01b = "[S01b] The variables declaration" ++ msgS01footer
msgS01c = "[S01c] The orders declaration"    ++ msgS01footer
msgS01d = "[S01d] The internals declaration" ++ msgS01footer
msgS01e = "[S01e] The events declaration"    ++ msgS01footer

msgS01footer = " is either missing or appears several times."

--------------------------------------------------------------------------------
-- Rule S02 --
--------------

checkS02 :: Spectre -> Bool
checkS02 s = allSymbolsDeclared (map name $ clocks ++ vars) (map name syms)
    where clocks = let cs = findDeclaredClocks s
                   in if cs == Nothing then [] else fromJust cs
          vars   = let vs = findDeclaredVariables s
                   in if vs == Nothing then [] else fromJust vs
          syms   = getAllSymbolsFromPredicates s
          name (Identifier n l c) = n
                            
ruleS02 = SemanticCheck checkS02 msgS02

msgS02 = "[S02] There is an undeclared symbol in a predicate."                            

--------------------------------------------------------------------------------
-- Rule S03 --
--------------

checkS03 :: Spectre -> Bool
checkS03 s = allSymbolsDeclared (map name $ clocks ++ vars) (map name syms)
    where clocks = let cs = findDeclaredClocks s
                   in if cs == Nothing then [] else fromJust cs
          vars   = let vs = findDeclaredVariables s
                   in if vs == Nothing then [] else fromJust vs
          syms   = getAllSymbolsFromAssignments s
          name (Identifier n _ _) = n
                        
ruleS03 = SemanticCheck checkS03 msgS03

msgS03 = "[S03] There is an undeclared symbol in an assignment."                            

--------------------------------------------------------------------------------
-- Rule S04 --
--------------

checkS04 :: Spectre -> Bool
checkS04 s = let (Initially (Identifier n _ _) _) = getInitially s
                 (Locations ls) = getLocations s
    in check n ls
    where check str [] = False
          check str ((Location (Identifier n _ _) _):xs) = if str == n then True
                                                           else check str xs

msgS04 = "[S04] The initial location is not declared as a location."

ruleS04 = SemanticCheck checkS04 msgS04

--------------------------------------------------------------------------------
-- Rule S05 --
--------------

checkS05 :: Spectre -> Bool
checkS05 s = let dl = getAllDeclaredLabels s
                 ts = getAllTransitions s
    in allLabelsDeclared (map name dl) (map name $ getLabels ts)
    where getLabels :: [Transition] -> [Identifier]
          getLabels [] = []
          getLabels ((Transition _ (Label i) _ _):xs) = [i] ++ (getLabels xs)
          getLabels ((Transition _ (None)    _ _):xs) = getLabels xs
          name :: Identifier -> String
          name (Identifier n _ _) = n
    
msgS05 = "[S05] There is an undeclared label used in one of the transitions."

ruleS05 = SemanticCheck checkS05 msgS05                 

--------------------------------------------------------------------------------
-- Rule S06 --
--------------

checkS06 :: Spectre -> Bool
checkS06 s = let dl = getAllDeclaredLocations s
                 ts = getAllTransitions s
    in allLocationsDeclared (map name dl) (map getLocation ts)
    where getLocation (Transition _ _ _ (Identifier n _ _)) = n
          name :: Identifier -> String
          name (Identifier n _ _) = n
    
msgS06 = "[S06] There is an undeclared location name as destination in one " ++
         "of the transitions."
         
ruleS06 = SemanticCheck checkS06 msgS06         

--------------------------------------------------------------------------------
-- Rule S07 --
--------------

checkS07 :: Spectre -> Bool
checkS07 spectre = numberOfDeclaredIds == numberOfUniqueIds
    where numberOfDeclaredIds = sum $ map length declared
          numberOfUniqueIds = length $ (nub.flat) declared
          declared = [c] ++ [v] ++ [o] ++ [e] ++ [i] ++ [l]
          get x = if x == Nothing then [] else fromJust x
          c = let c' = findDeclaredClocks spectre in map name $ get c'
          v = let v' = findDeclaredVariables spectre in map name $ get v'
          o = let o' = findDeclaredOrders spectre in map name $ get o'
          e = let e' = findDeclaredEvents spectre in map name $ get e'
          i = let i' = findDeclaredInternals spectre in map name $ get i'
          l = map name $ getAllDeclaredLocations spectre
          flat [] = []
          flat (x:xs) = x `union` (flat xs)
          name (Identifier n _ _) = n

msgS07 = "[S07] There is an identifier which is defined more than once."

ruleS07 = SemanticCheck checkS07 msgS07          

--------------------------------------------------------------------------------
-- Rule S09 --
--------------

checkS09 :: Spectre -> Bool
checkS09 s = let ps = getAllPredicates s in doCheck ps
    where doCheck :: [Predicate] -> Bool
          doCheck ps = not $ False `elem` (map (isValidPredicate s) ps)
          
msgS09 = "[S09] There is a predicate which either contains both variables and"++
         " clocks, or contains neither."
         
ruleS09 = SemanticCheck checkS09 msgS09
                   
--------------------------------------------------------------------------------
-- Rule D01 --
--------------
 
countOrderDecos :: Spectre -> String -> Int
countOrderDecos s order = let d = getAllDecorationItems s in doCount d
    where doCount [] = 0
          doCount ((Order o _ _):xs) | (name o) == order = 1 + (doCount xs)
                                   | otherwise = doCount xs
          doCount (_:xs) = doCount xs
          name :: Identifier -> String
          name (Identifier n _ _) = n                                     
 
countEventDecos :: Spectre -> String -> Int
countEventDecos s event = let d = getAllDecorationItems s in doCount d
    where doCount [] = 0
          doCount ((Event e _ _ _):xs) | (name e) == event = 1 + (doCount xs)
                                   | otherwise = doCount xs
          doCount (_:xs) = doCount xs
          name :: Identifier -> String
          name (Identifier n _ _) = n   
          
countInternalDecos :: Spectre -> String -> Int
countInternalDecos s int = let d = getAllDecorationItems s in doCount d
    where doCount [] = 0
          doCount ((Internal i _ _):xs) | (name i) == int = 1 + (doCount xs)
                                      | otherwise = doCount xs
          doCount (_:xs) = doCount xs 
          name :: Identifier -> String
          name (Identifier n _ _) = n   

okOrders :: Spectre -> Bool          
okOrders s = let declared = findDeclaredOrders s
    in if declared == Nothing then True else okOrders' s (fromJust declared)
    where okOrders' :: Spectre -> [Identifier] -> Bool
          okOrders' s [] = True
          okOrders' s ((Identifier n _ _):xs) = if ok s n then okOrders' s xs
                                                else False
          ok s n = countOrderDecos s n == 1

okEvents:: Spectre -> Bool          
okEvents s = let declared = findDeclaredEvents s
    in if declared == Nothing then True else okEvents' s (fromJust declared)
    where okEvents' :: Spectre -> [Identifier] -> Bool
          okEvents' s [] = True
          okEvents' s ((Identifier n _ _):xs) = if ok s n then okEvents' s xs
                                                else False
          ok s n = countEventDecos s n == 1

okInternals :: Spectre -> Bool          
okInternals s = let declared = findDeclaredInternals s
    in if declared == Nothing then True else okInterns' s (fromJust declared)
    where okInterns' :: Spectre -> [Identifier] -> Bool
          okInterns' s [] = True
          okInterns' s ((Identifier n _ _):xs) = if ok s n then okInterns' s xs
                                                else False
          ok s n = countInternalDecos s n == 1


checkD01a = okEvents
checkD01b = okInternals
checkD01c = okOrders

msgD01header = "[D01] *Exactly one* decoration item must be provided for each "

msgD01a = msgD01header ++ "event label." 
msgD01b = msgD01header ++ "internal label." 
msgD01c = msgD01header ++ "order label." 

ruleD01a = SemanticCheck checkD01a msgD01a
ruleD01b = SemanticCheck checkD01b msgD01b
ruleD01c = SemanticCheck checkD01c msgD01c

rulesD01 = [ruleD01a, ruleD01b, ruleD01c]
              
--------------------------------------------------------------------------------
-- Rule D02 --
--------------

checkD02a s = allSymbolsDeclared (map name $ clocks ++ vars) (map name syms)
    where clocks = let cs = findDeclaredClocks s
                   in if cs == Nothing then [] else fromJust cs
          vars   = let vs = findDeclaredVariables s
                   in if vs == Nothing then [] else fromJust vs
          syms   = getAllReadSymbols s
          name (Identifier n l c) = n
          
checkD02b s = allSymbolsDeclared (map name $ clocks ++ vars) (map name syms)
    where clocks = let cs = findDeclaredClocks s
                   in if cs == Nothing then [] else fromJust cs
          vars   = let vs = findDeclaredVariables s
                   in if vs == Nothing then [] else fromJust vs
          syms   = getAllWrittenSymbols s
          name (Identifier n l c) = n          

msgD02a = "[D02a] A 'reading' decoration uses an undeclared identifier."
msgD02b = "[D02b] A 'writing' decoration uses an undeclared identifier."

ruleD02a = SemanticCheck checkD02a msgD02a
ruleD02b = SemanticCheck checkD02b msgD02b

rulesD02 = [ruleD02a, ruleD02b]

--------------------------------------------------------------------------------
-- Rule D03 --
--------------

checkD03 spectre = intersect namespace parameters == []
    where namespace = (nub.flat) declared
          declared = [c] ++ [v] ++ [o] ++ [e] ++ [i] ++ [l]
          parameters = map name $ getAllWritingParameters spectre
          get x = if x == Nothing then [] else fromJust x
          c = let c' = findDeclaredClocks spectre in map name $ get c'
          v = let v' = findDeclaredVariables spectre in map name $ get v'
          o = let o' = findDeclaredOrders spectre in map name $ get o'
          e = let e' = findDeclaredEvents spectre in map name $ get e'
          i = let i' = findDeclaredInternals spectre in map name $ get i'
          l = map name $ getAllDeclaredLocations spectre
          flat [] = []
          flat (x:xs) = x `union` (flat xs)
          name (Identifier n _ _) = n
          
msgD03 = "[D03] A parameter for a 'writing' decoration cannot be " ++ 
         "declared elsewhere."          

ruleD03 = SemanticCheck checkD03 msgD03

--------------------------------------------------------------------------------
-- Rule D04 --
--------------

allHaveOneReading :: Spectre -> [Identifier] -> Bool
allHaveOneReading s [] = True
allHaveOneReading s (x:xs) = ok s x && (allHaveOneReading s xs)
    where ok s x = (countReadingsForVariable s x) == 1

checkD04 :: Spectre -> Bool
checkD04 s = allHaveOneReading s $ removeClocks s (a ++ b)
    where a = getAllSymbolsFromPredicates s
          b = getAllRValuesFromAssignments s
          removeClocks :: Spectre -> [Identifier] -> [Identifier]
          removeClocks s ids = let cs = findDeclaredClocks s in doRemove cs ids
          doRemove :: Maybe [Identifier] -> [Identifier] -> [Identifier]
          doRemove Nothing a = a
          doRemove (Just []) ids = ids
          doRemove (Just (c:cs)) ids = doRemove (Just cs) (doRemove' c ids)
          doRemove' :: Identifier -> [Identifier] -> [Identifier]
          doRemove' _ [] = []
          doRemove' c@(Identifier n1 _ _) (i@(Identifier n2 _ _):xs) = result
                where result = if n1 == n2 then doRemove' c xs
                               else (doRemove' c xs) ++ [i]                                    

ruleD04 = SemanticCheck checkD04 msgD04

msgD04 = "[D04] A variable was found in an expression and has not *exactly " ++
         "one* 'reading' decoration."

--------------------------------------------------------------------------------
-- Rule D06 --
--------------

allHaveOneWriting :: Spectre -> [Identifier] -> Bool
allHaveOneWriting s [] = True
allHaveOneWriting s (x:xs) = ok s x && (allHaveOneWriting s xs)
    where ok s x = (countWritingsForVariable s x) == 1

checkD06 :: Spectre -> Bool
checkD06 s = allHaveOneWriting s $ removeClocks s (a)
    where a = getAllLValuesFromAssignments s
          removeClocks :: Spectre -> [Identifier] -> [Identifier]
          removeClocks s ids = let cs = findDeclaredClocks s in doRemove cs ids
          doRemove :: Maybe [Identifier] -> [Identifier] -> [Identifier]
          doRemove Nothing a = a
          doRemove (Just []) ids = ids
          doRemove (Just (c:cs)) ids = doRemove (Just cs) (doRemove' c ids)
          doRemove' :: Identifier -> [Identifier] -> [Identifier]
          doRemove' _ [] = []
          doRemove' c@(Identifier n1 _ _) (i@(Identifier n2 _ _):xs) = result
                where result = if n1 == n2 then doRemove' c xs
                               else (doRemove' c xs) ++ [i]                                    

ruleD06 = SemanticCheck checkD06 msgD06

msgD06 = "[D06] A variable was found in the left part of an assignment and " ++
         "has not *exactly one* 'writing' decoration."

--------------------------------------------------------------------------------
-- Rule D08 --
--------------

checkD08 :: Spectre -> Bool
checkD08 s = loc2loc && loc2any && any2loc
    where loc2loc = let ids = getAllFromLocToLoc s in doTest2 ids
          any2loc = let ids = getAllFromAnyToLoc s in doTest ids
          loc2any = let ids = getAllFromLocToAny s in doTest ids
          doTest2 :: [(Identifier,Identifier)] -> Bool
          doTest2 [] = True
          doTest2 ((a,b):xs) = (declaredAsLocation s (name a)) &&
                               (declaredAsLocation s (name b)) && (doTest2 xs)
          doTest :: [Identifier] -> Bool
          doTest [] = True
          doTest (a:xs) = (declaredAsLocation s (name a)) && (doTest xs)                               
          name :: Identifier -> String
          name (Identifier n _ _) = n
          
msgD08 = "[D08] A transition decoration refers to an undeclared location."                    
          
ruleD08 = SemanticCheck checkD08 msgD08          

--------------------------------------------------------------------------------
-- Rule D09 --
--------------

checkD09 :: Spectre -> Bool
checkD09 s = loc2loc && loc2any && any2loc
    where loc2loc = let ids = getAllRestrictFromLocToLoc s in doTest2 ids
          any2loc = let ids = getAllRestrictFromAnyToLoc s in doTest ids
          loc2any = let ids = getAllRestrictFromLocToAny s in doTest ids
          doTest2 :: [(Identifier,Identifier)] -> Bool
          doTest2 [] = True
          doTest2 ((a,b):xs) = (declaredAsLocation s (name a)) &&
                               (declaredAsLocation s (name b)) && (doTest2 xs)
          doTest :: [Identifier] -> Bool
          doTest [] = True
          doTest (a:xs) = (declaredAsLocation s (name a)) && (doTest xs)                               
          name :: Identifier -> String
          name (Identifier n _ _) = n
          
msgD09 = "[D09] A restriction decoration refers to an undeclared location."                    
          
ruleD09 = SemanticCheck checkD09 msgD09

--------------------------------------------------------------------------------
-- Rule D10 --
--------------

checkD10a :: Spectre -> Bool
checkD10a s = (countGlobals s) == 1

checkD10b :: Spectre -> Bool
checkD10b s = (countStartups s) == 1

checkD10c :: Spectre -> Bool
checkD10c s = (countCleanups s) == 1
         
msgD10a = "[D10a] There must be *exactly one* 'global' decoration."
msgD10b = "[D10b] There must be *exactly one* 'startup' decoration."          
msgD10c = "[D10c] There must be *exactly one* 'cleanup' decoration."                    

ruleD10a = SemanticCheck checkD10a msgD10a
ruleD10b = SemanticCheck checkD10b msgD10b
ruleD10c = SemanticCheck checkD10c msgD10c

rulesD10 = [ruleD10a, ruleD10b, ruleD10c]

--------------------------------------------------------------------------------
-- Rule D11 --
--------------

checkD11a :: Spectre -> Bool
checkD11a s = loc2loc && loc2any && any2loc && any2any
    where loc2loc = let ids = getAllFromLocToLoc s in doTest $ map name2 ids
          any2loc = let ids = getAllFromAnyToLoc s in doTest $ map name ids
          loc2any = let ids = getAllFromLocToAny s in doTest $ map name ids        
          any2any = countFromAnyToAny s <= 1
          doTest :: Eq a => [a] -> Bool
          doTest [] = True
          doTest (a:xs) = not (a `elem` xs) && (doTest xs)
          name :: Identifier -> String
          name (Identifier n _ _) = n
          name2 :: (Identifier,Identifier) -> (String,String)
          name2 ((Identifier n1 _ _),(Identifier n2 _ _)) = (n1,n2)

checkD11b :: Spectre -> Bool
checkD11b s = loc2loc && loc2any && any2loc && any2any
    where loc2loc = let ids = getAllRestrictFromLocToLoc s
                    in doTest $ map name2 ids
          any2loc = let ids = getAllRestrictFromAnyToLoc s
                    in doTest $ map name ids
          loc2any = let ids = getAllRestrictFromLocToAny s
                    in doTest $ map name ids
          any2any = countRestrictFromAnyToAny s <= 1
          doTest :: Eq a => [a] -> Bool
          doTest [] = True
          doTest (a:xs) = not (a `elem` xs) && (doTest xs)
          name :: Identifier -> String
          name (Identifier n _ _) = n
          name2 :: (Identifier,Identifier) -> (String,String)
          name2 ((Identifier n1 _ _),(Identifier n2 _ _)) = (n1,n2)
          
msgD11a = "[D11a] A transition is decorated more than once."
msgD11b = "[D11a] A transition restriction is decorated more than once."

ruleD11a = SemanticCheck checkD11a msgD11a
ruleD11b = SemanticCheck checkD11b msgD11b

rulesD11 = [ruleD11a] ++ [ruleD11b]

--------------------------------------------------------------------------------

errorChecks :: [SemanticCheck]
errorChecks = rulesS01 ++ [ruleS02] ++ [ruleS03] ++ [ruleS04] ++ [ruleS05] ++
              [ruleS06] ++ [ruleS07] ++ rulesD01 ++ rulesD02 ++ [ruleD03] ++ 
              [ruleD04] ++ [ruleD06] ++ [ruleD08] ++ [ruleD09] ++ rulesD10 ++
              rulesD11 ++ [ruleS09]

warningChecks :: [SemanticCheck]
warningChecks = []

--------------------------------------------------------------------------------

{-
    This is the checking machinery-code. It's ugly, but it works.
    What it does is simple : it runs through the lists 'errorChecks' and
    'warningChecks' defined above and runs each semantic check found on the 
    Spectre syntax tree. Everytime a check fails, a warning or error is appended
    to a CheckOutput data structure. At the end of the run, the errors and 
    warning messages are displayed and the program exits with a success code 
    (returning zero) only if the program did not issue any error messages.
    In the other case, it returns 1.
-}

data CheckerOutput = CheckerOutput [String] -- warnings
                                   [String] -- errors

showWarnings :: CheckerOutput -> IO ()
showWarnings (CheckerOutput [] e) = putStr ""
showWarnings (CheckerOutput (w:ws) e) = 
    do hPutStrLn stderr ("Semantic warning : " ++ w)
       showWarnings (CheckerOutput ws e)

showErrors :: CheckerOutput -> IO ()
showErrors (CheckerOutput w []) = putStr ""
showErrors (CheckerOutput w (e:es)) = 
    do hPutStrLn stderr ("Semantic error : " ++ e)
       showErrors (CheckerOutput w es)

exitCode(CheckerOutput w []) = ExitSuccess
exitCode(CheckerOutput w _ ) = (ExitFailure 1)

doWarningCheck :: CheckerOutput -> Spectre -> SemanticCheck -> CheckerOutput
doWarningCheck out@(CheckerOutput w e) spectre (SemanticCheck check msg) =
    if check spectre then out else (CheckerOutput (msg:w) e)

doErrorCheck :: CheckerOutput -> Spectre -> SemanticCheck -> CheckerOutput
doErrorCheck out@(CheckerOutput w e) spectre (SemanticCheck check msg) =
    if check spectre then out else (CheckerOutput w (msg:e))

doWarningChecks :: CheckerOutput -> Spectre -> CheckerOutput
doWarningChecks out s = doWarningChecks' out s warningChecks

doWarningChecks' :: CheckerOutput -> Spectre -> [SemanticCheck] -> CheckerOutput 
doWarningChecks' out s [] = out
doWarningChecks' out s (x:xs) = let out' = doWarningChecks' out s xs
                                in doWarningCheck out' s x

doErrorChecks :: CheckerOutput -> Spectre -> CheckerOutput
doErrorChecks out s = doErrorChecks' out s errorChecks

doErrorChecks' :: CheckerOutput -> Spectre -> [SemanticCheck] -> CheckerOutput 
doErrorChecks' out s [] = out
doErrorChecks' out s (x:xs) = let out' = doErrorChecks' out s xs
                              in doErrorCheck out' s x

checkSemantics :: Spectre -> IO CheckerOutput
checkSemantics s = return (out)
    where out = doErrorChecks warn s
          init = CheckerOutput [] []
          warn = doWarningChecks init s
          
--------------------------------------------------------------------------------          
          
readSpectre :: String -> Spectre
readSpectre s = read s

main = do str <- getContents
          out <- checkSemantics (readSpectre str)
          showWarnings(out)
          showErrors(out)
          if exitCode(out) == ExitSuccess then do (putStr str) 
                                          else do hPutStr stderr ""
          exitWith (exitCode(out))
          
