module SyntaxTree where

import Maybe
import List

data Spectre = Spectre Specification
                       Decoration                         deriving(Eq,Show,Read)

data Specification = Specification Identifier
                                   Declarations           
                                   Initially              
                                   Locations              deriving(Eq,Show,Read)
                   
data Decoration = Decoration Identifier
                             DecorationItems              deriving(Eq,Show,Read)
                             
data DecorationItems = DecorationItems [DecorationItem]   deriving(Eq,Show,Read)      

data DecorationItem = Global Code
                    | Startup Code
                    | Cleanup Code
                    | Order Identifier Code Code
                    | Event Identifier Code Code Code
                    | Internal Identifier Code Code
                    | Reading Identifier Code
                    | Writing Identifier Identifier Code
                    | FromLocToLoc Identifier Identifier Code Code
                    | FromAnyToLoc Identifier Code Code
                    | FromLocToAny Identifier Code Code
                    | FromAnyToAny Code Code
                    | Restriction RestrictionHeader Code  deriving(Eq,Show,Read)
               
data Code = Code String
          | Nop                                           deriving(Eq,Show,Read)
                                     
data RestrictionHeader = RestrictFromLocToLoc Identifier Identifier
                       | RestrictFromAnyToLoc Identifier
                       | RestrictFromLocToAny Identifier
                       | RestrictFromAnyToAny
                                                          deriving(Eq,Show,Read)
                             
data Declarations = Declarations [Declaration]            deriving(Eq,Show,Read)
data Declaration = ClocksDeclaration [Identifier]
                 | OrdersDeclaration [Identifier]
                 | EventsDeclaration [Identifier]
                 | InternalsDeclaration [Identifier]
                 | VarsDeclaration [Identifier]           deriving(Eq,Show,Read)
                 
data Expression = Addition Expression Term
                | Substraction Expression Term
                | Term Term                               deriving(Eq,Show,Read)
                
data Term = Multiplication Term Factor
          | Division Term Factor
          | Factor Factor                                 deriving(Eq,Show,Read)
          
data Factor = Number Int
            | Symbol Identifier
            | Parenthesis Expression                      deriving(Eq,Show,Read)
                      
data Initially = Initially Identifier
                           Assignments                    deriving(Eq,Show,Read)

data Assignments = Assignments [Assignment]               deriving(Eq,Show,Read)

data Assignment = Assignment Identifier Expression        deriving(Eq,Show,Read)

data Locations = Locations [Location]                     deriving(Eq,Show,Read)

data Location = Location Identifier Transitions           deriving(Eq,Show,Read)

data Transitions = Transitions [Transition]               deriving(Eq,Show,Read)

data Transition = Transition Predicates
                             Label
                             Assignments
                             Identifier                   deriving(Eq,Show,Read)
                             
data Label = Label Identifier
           | None                                         deriving(Eq,Show,Read)
                           
data Predicates = Predicates [Predicate]                  deriving(Eq,Show,Read)

data Predicate = Predicate Expression
                           Comparator
                           Expression                     deriving(Eq,Show,Read)
                           
data Comparator = EqualTo
                | SmallerThan
                | LargerThan                              deriving(Eq,Show,Read)

data Identifier = Identifier String Int Int               deriving(Eq,Show,Read)
                
--------------------------------------------------------------------------------

getSpecification :: Spectre -> Specification
getSpecification (Spectre s d) = s

getDecoration :: Spectre -> Decoration
getDecoration    (Spectre s d) = d

getDecorationName :: Spectre -> Identifier
getDecorationName s = let (Decoration id _) = getDecoration s in id

getDeclarations :: Spectre -> Declarations
getDeclarations s = let (Specification _ d _ _) = getSpecification s in d

getInitially :: Spectre -> Initially
getInitially s = let (Specification _ _ i _) = getSpecification s in i

getLocations :: Spectre -> Locations
getLocations s = let (Specification _ _ _ l) = getSpecification s in l

getAllLocations :: Spectre -> [Location]
getAllLocations s = let (Locations ls) = getLocations s in ls

getDecorationItems :: Spectre -> DecorationItems
getDecorationItems s = let (Decoration _ i) = getDecoration s in i

findClocksDeclaration :: Spectre -> Maybe Declaration
findClocksDeclaration s = let (Declarations d) = getDeclarations s
    in doFind d
    where doFind [] = Nothing
          doFind (c@(ClocksDeclaration _):_) = Just c
          doFind (_:xs) = doFind xs

findOrdersDeclaration :: Spectre -> Maybe Declaration
findOrdersDeclaration s = let (Declarations d) = getDeclarations s
    in doFind d
    where doFind [] = Nothing
          doFind (c@(OrdersDeclaration _):_) = Just c
          doFind (_:xs) = doFind xs                       

findEventsDeclaration :: Spectre -> Maybe Declaration                        
findEventsDeclaration s = let (Declarations d) = getDeclarations s
    in doFind d
    where doFind [] = Nothing
          doFind (c@(EventsDeclaration _):_) = Just c
          doFind (_:xs) = doFind xs                       

findInternalsDeclaration :: Spectre -> Maybe Declaration
findInternalsDeclaration s = let (Declarations d) = getDeclarations s
    in doFind d
    where doFind [] = Nothing
          doFind (c@(InternalsDeclaration _):_) = Just c
          doFind (_:xs) = doFind xs

findVarsDeclaration :: Spectre -> Maybe Declaration                  
findVarsDeclaration s = let (Declarations d) = getDeclarations s
    in doFind d
    where doFind [] = Nothing
          doFind (c@(VarsDeclaration _):_) = Just c
          doFind (_:xs) = doFind xs                                

countClocksDeclaration :: Spectre -> Int                        
countClocksDeclaration s = let (Declarations d) = getDeclarations s
    in doCount d
    where doCount [] = 0
          doCount (c@(ClocksDeclaration _):xs) = 1 + doCount xs
          doCount (_:xs) = doCount xs

countOrdersDeclaration :: Spectre -> Int                        
countOrdersDeclaration s = let (Declarations d) = getDeclarations s
    in doCount d
    where doCount [] = 0
          doCount (c@(OrdersDeclaration _):xs) = 1 + doCount xs
          doCount (_:xs) = doCount xs

countEventsDeclaration :: Spectre -> Int          
countEventsDeclaration s = let (Declarations d) = getDeclarations s
    in doCount d
    where doCount [] = 0
          doCount (c@(EventsDeclaration _):xs) = 1 + doCount xs
          doCount (_:xs) = doCount xs

countInternalsDeclaration :: Spectre -> Int                  
countInternalsDeclaration s = let (Declarations d) = getDeclarations s
    in doCount d
    where doCount [] = 0
          doCount (c@(InternalsDeclaration _):xs) = 1 + doCount xs
          doCount (_:xs) = doCount xs

countVarsDeclaration :: Spectre -> Int          
countVarsDeclaration s = let (Declarations d) = getDeclarations s
    in doCount d
    where doCount [] = 0
          doCount (c@(VarsDeclaration _):xs) = 1 + doCount xs
          doCount (_:xs) = doCount xs

getAllTransitions :: Spectre -> [Transition]
getAllTransitions s = let (Locations ls) = getLocations s
    in doGet ls
    where doGet [] = []
          doGet ((Location _ (Transitions t)):xs) = t ++ doGet xs

getAllLocTransitions :: Spectre -> [(String, Transition)]
getAllLocTransitions s = let (Locations ls) = getLocations s
    in doGet ls
    where doGet [] = []
          doGet ((Location id (Transitions t)):xs) = (mk id t) ++ doGet xs
          mk :: Identifier -> [Transition] -> [(String, Transition)]
          mk _ [] = []
          mk i@(Identifier n _ _) (x:xs) = [(n,x)] ++ (mk i xs)

getTransitionNumber :: Spectre -> Transition -> String -> Int
getTransitionNumber s t a = let ls = getAllLocations s in doGet ls
    where doGet :: [Location] -> Int
          doGet [] = error "getTransitionNumber : not found !"
          doGet ((Location (Identifier n _ _) ts):xs) = if a == n 
                                                        then doGet' 1 (p ts)
                                                        else doGet xs
          p :: Transitions -> [Transition]
          p (Transitions xs) =  xs 
          doGet' :: Int -> [Transition] -> Int
          doGet' i [] = error "getTransitionNumber : not found !"
          doGet' i (x:xs) = if x == t then i else doGet' (i+1) xs
         
getAllPredicates :: Spectre -> [Predicate]          
getAllPredicates s = let ts = getAllTransitions s
    in doGet ts
    where doGet [] = []
          doGet ((Transition (Predicates p) _ _ _):xs) = p ++ doGet xs          

getAllSymbolsE :: Expression -> [Identifier]          
getAllSymbolsE (Addition e t)     = (getAllSymbolsE e) ++ (getAllSymbolsT t)
getAllSymbolsE (Substraction e t) = (getAllSymbolsE e) ++ (getAllSymbolsT t)
getAllSymbolsE (Term t) = getAllSymbolsT t

getAllSymbolsT :: Term -> [Identifier]
getAllSymbolsT (Division t f)       = (getAllSymbolsT t) ++ (getAllSymbolsF f)
getAllSymbolsT (Multiplication t f) = (getAllSymbolsT t) ++ (getAllSymbolsF f)                        
getAllSymbolsT (Factor f) = getAllSymbolsF f

getAllSymbolsF :: Factor -> [Identifier]
getAllSymbolsF (Number i) = []
getAllSymbolsF (Symbol s) = [s]
getAllSymbolsF (Parenthesis e) = getAllSymbolsE e                        

findDeclaredClocks :: Spectre -> Maybe [Identifier]                        
findDeclaredClocks s = if cd == Nothing then Nothing else clocks
    where cd = findClocksDeclaration s
          getClocks (ClocksDeclaration cs) = cs
          clocks = Just (getClocks (fromJust cd))

findDeclaredVariables :: Spectre -> Maybe [Identifier]                        
findDeclaredVariables s = if vd == Nothing then Nothing else vars
    where vd = findVarsDeclaration s
          getVars (VarsDeclaration vs) = vs
          vars = Just (getVars (fromJust vd))                        

findDeclaredEvents :: Spectre -> Maybe [Identifier]
findDeclaredEvents s = if ed == Nothing then Nothing else events
    where ed = findEventsDeclaration s
          getEvents (EventsDeclaration es) = es
          events = Just (getEvents (fromJust ed))  

findDeclaredOrders :: Spectre -> Maybe [Identifier]
findDeclaredOrders s = if od == Nothing then Nothing else orders
    where od = findOrdersDeclaration s
          getOrders (OrdersDeclaration os) = os
          orders = Just (getOrders (fromJust od))  

findDeclaredInternals :: Spectre -> Maybe [Identifier]
findDeclaredInternals s = if id == Nothing then Nothing else internals
    where id = findInternalsDeclaration s
          getInternals (InternalsDeclaration is) = is
          internals = Just (getInternals (fromJust id))  

getAllAssignments :: Spectre -> [Assignment]                        
getAllAssignments s = let (Initially _ (Assignments a)) = getInitially s
                          ts = getAllTransitions s
    in a ++ (doGet ts)
    where doGet [] = []
          doGet ((Transition _ _ (Assignments a) _):xs) = a ++ (doGet xs)

getAllDeclaredLabels :: Spectre -> [Identifier]                                               
getAllDeclaredLabels s = e ++ i ++ o
    where e = let d = findDeclaredEvents s in get d
          i = let d = findDeclaredInternals s in get d
          o = let d = findDeclaredOrders s in get d
          get x = if x == Nothing then [] else fromJust x
          
getAllDeclaredLocations :: Spectre -> [Identifier]          
getAllDeclaredLocations s = let (Locations ls) = getLocations s in get ls
    where get [] = []
          get ((Location i _):xs) = [i] ++ get xs                                                
                        
getAllDecorationItems :: Spectre -> [DecorationItem]
getAllDecorationItems s = let (DecorationItems di) = getDecorationItems s in di
                        
getAllReadSymbols :: Spectre -> [Identifier]
getAllReadSymbols s = let di = getAllDecorationItems s in f di
    where f :: [DecorationItem] -> [Identifier]
          f [] = []
          f ((Reading i _):xs) = [i] ++ (f xs)
          f (_:xs) = f xs
          
getAllWrittenSymbols :: Spectre -> [Identifier]
getAllWrittenSymbols s = let di = getAllDecorationItems s in f di
    where f :: [DecorationItem] -> [Identifier]
          f [] = []
          f ((Writing i _ _):xs) = [i] ++ (f xs)
          f (_:xs) = f xs
          
getAllWritingParameters :: Spectre -> [Identifier]
getAllWritingParameters s = let di = getAllDecorationItems s in f di
    where f :: [DecorationItem] -> [Identifier]
          f [] = []
          f ((Writing _ p _):xs) = [p] ++ (f xs)
          f (_:xs) = f xs
                                                
getAllSymbolsFromAssignments :: Spectre -> [Identifier]
getAllSymbolsFromAssignments s = let ps = getAllAssignments s
    in doGet ps
    where doGet [] = []
          doGet ((Assignment s e):xs) = [s] ++ (getAllSymbolsE e) ++ (doGet xs)


getAllLValuesFromAssignments :: Spectre -> [Identifier]
getAllLValuesFromAssignments s = let ps = getAllAssignments s
    in doGet ps
    where doGet [] = []
          doGet ((Assignment s _):xs) = [s] ++ (doGet xs)

getAllRValuesFromAssignments :: Spectre -> [Identifier]
getAllRValuesFromAssignments s = let ps = getAllAssignments s
    in doGet ps
    where doGet [] = []
          doGet ((Assignment _ e):xs) = (getAllSymbolsE e) ++ (doGet xs)

getAllSymbolsFromPredicates :: Spectre -> [Identifier]
getAllSymbolsFromPredicates s = let ps = getAllPredicates s
    in doGet ps
    where doGet [] = []
          doGet ((Predicate e1 c e2):xs) = (g1 e1) ++ (g1 e2) ++ (doGet xs)
          g1 = getAllSymbolsE                        
                        
                        
countReadingsForVariable :: Spectre -> Identifier -> Int
countReadingsForVariable s i1 = let ds = getAllDecorationItems s in doCount ds
    where doCount :: [DecorationItem] -> Int
          doCount [] = 0
          doCount ((Reading i2 _):xs) = (ifSame i1 i2) + (doCount xs)
          doCount (_:xs) = doCount xs
          ifSame :: Identifier -> Identifier -> Int
          ifSame (Identifier n1 _ _) (Identifier n2 _ _) = if n1 == n2 then 1
                                                           else 0
                        
countWritingsForVariable :: Spectre -> Identifier -> Int
countWritingsForVariable s i1 = let ds = getAllDecorationItems s in doCount ds
    where doCount :: [DecorationItem] -> Int
          doCount [] = 0
          doCount ((Writing i2 _ _):xs) = (ifSame i1 i2) + (doCount xs)
          doCount (_:xs) = doCount xs
          ifSame :: Identifier -> Identifier -> Int
          ifSame (Identifier n1 _ _) (Identifier n2 _ _) = if n1 == n2 then 1
                                                           else 0                        

declaredAsLocation :: Spectre -> String -> Bool
declaredAsLocation s n = let dl = getAllDeclaredLocations s in doTest n dl
    where doTest :: String -> [Identifier] -> Bool
          doTest s [] = False
          doTest s ((Identifier n _ _):xs) = if n == s then True
                                             else doTest s xs                    

getAllFromLocToLoc :: Spectre -> [(Identifier, Identifier)]
getAllFromLocToLoc s = let di = getAllDecorationItems s in doGet di
    where doGet :: [DecorationItem] -> [(Identifier, Identifier)]
          doGet [] = []
          doGet ((FromLocToLoc a b _ _):xs) = (a,b) : (doGet xs)
          doGet (_:xs) = doGet xs
          
getAllFromAnyToLoc :: Spectre -> [Identifier]
getAllFromAnyToLoc s = let di = getAllDecorationItems s in doGet di
    where doGet :: [DecorationItem] -> [Identifier]
          doGet [] = []
          doGet ((FromAnyToLoc a _ _):xs) = a : (doGet xs)
          doGet (_:xs) = doGet xs
          
getAllFromLocToAny :: Spectre -> [Identifier]
getAllFromLocToAny s = let di = getAllDecorationItems s in doGet di
    where doGet :: [DecorationItem] -> [Identifier]
          doGet [] = []
          doGet ((FromLocToAny a _ _):xs) = a : (doGet xs)
          doGet (_:xs) = doGet xs
                                  
getFromAnyToAny :: Spectre -> Maybe DecorationItem
getFromAnyToAny s = let di = getAllDecorationItems s in doGet di
    where doGet :: [DecorationItem] -> Maybe DecorationItem
          doGet [] = Nothing
          doGet (f@(FromAnyToAny _ _):xs) = Just f
          doGet (_:xs) = doGet xs
                    
countFromAnyToAny :: Spectre -> Int
countFromAnyToAny s = let di = getAllDecorationItems s in doCount di
    where doCount :: [DecorationItem] -> Int
          doCount [] = 0
          doCount ((FromAnyToAny _ _):xs) = 1 + (doCount xs)
          doCount (_:xs) = doCount xs
          
getAllRestrictionHeaders :: Spectre -> [RestrictionHeader]
getAllRestrictionHeaders s = let di = getAllDecorationItems s in doGet di
    where doGet :: [DecorationItem] -> [RestrictionHeader]
          doGet [] = []
          doGet ((Restriction h _):xs) = [h] ++ (doGet xs)
          doGet (_:xs) = doGet xs

getAllRestrictFromLocToLoc :: Spectre -> [(Identifier,Identifier)]
getAllRestrictFromLocToLoc s = let hs = getAllRestrictionHeaders s in doGet hs
    where doGet :: [RestrictionHeader] -> [(Identifier,Identifier)]
          doGet [] = []
          doGet ((RestrictFromLocToLoc a b):xs) = [(a,b)] ++ (doGet xs)
          doGet (_:xs) = doGet xs
          
getAllRestrictFromAnyToLoc :: Spectre -> [Identifier]
getAllRestrictFromAnyToLoc s = let hs = getAllRestrictionHeaders s in doGet hs
    where doGet :: [RestrictionHeader] -> [Identifier]
          doGet [] = []
          doGet ((RestrictFromAnyToLoc a):xs) = [a] ++ (doGet xs)
          doGet (_:xs) = doGet xs

getAllRestrictFromLocToAny :: Spectre -> [Identifier]
getAllRestrictFromLocToAny s = let hs = getAllRestrictionHeaders s in doGet hs
    where doGet :: [RestrictionHeader] -> [Identifier]
          doGet [] = []
          doGet ((RestrictFromLocToAny a):xs) = [a] ++ (doGet xs)
          doGet (_:xs) = doGet xs
          
getRestrictFromAnyToAny :: Spectre -> Maybe DecorationItem
getRestrictFromAnyToAny s = let di = getAllDecorationItems s in doGet di
    where doGet :: [DecorationItem] -> Maybe DecorationItem
          doGet [] = Nothing
          doGet (r@(Restriction RestrictFromAnyToAny _):xs) = Just r
          doGet (_:xs) = doGet xs
          
countRestrictFromAnyToAny :: Spectre -> Int
countRestrictFromAnyToAny s = let di = getAllDecorationItems s in doCount di
    where doCount :: [DecorationItem] -> Int
          doCount [] = 0
          doCount ((Restriction RestrictFromAnyToAny _):xs) = 1 + (doCount xs)
          doCount (_:xs) = doCount xs  

countGlobals :: Spectre -> Int
countGlobals s = let di = getAllDecorationItems s in doCount di
    where doCount :: [DecorationItem] -> Int
          doCount [] = 0
          doCount ((Global _):xs) = 1 + (doCount xs)
          doCount (_:xs) = doCount xs
          
countCleanups :: Spectre -> Int
countCleanups s = let di = getAllDecorationItems s in doCount di
    where doCount :: [DecorationItem] -> Int
          doCount [] = 0
          doCount ((Cleanup _):xs) = 1 + (doCount xs)
          doCount (_:xs) = doCount xs

countStartups :: Spectre -> Int
countStartups s = let di = getAllDecorationItems s in doCount di
    where doCount :: [DecorationItem] -> Int
          doCount [] = 0
          doCount ((Startup _):xs) = 1 + (doCount xs)
          doCount (_:xs) = doCount xs

getGlobalDecoration :: Spectre -> DecorationItem
getGlobalDecoration s =  let di = getAllDecorationItems s in doGet di
    where doGet :: [DecorationItem] -> DecorationItem
          doGet [] = error "Could not find global decoration."
          doGet (g@(Global _):xs) = g
          doGet (_:xs) = doGet xs
          
isClockPredicate :: Spectre -> Predicate -> Bool
isClockPredicate s (Predicate e1 _ e2) = 
    let cs = proj1 $ findDeclaredClocks s
        vs = proj1 $ findDeclaredVariables s
        e1s = map proj2 $ getAllSymbolsE e1
        e2s = map proj2 $ getAllSymbolsE e2
    in ((intersect vs $ e1s `union` e2s) == []) &&
       ((intersect cs $ e1s `union` e2s) /= [])
    where proj1 :: Maybe [Identifier] -> [String]
          proj1 Nothing = []
          proj1 (Just x) = map proj2 x
          proj2 :: Identifier -> String
          proj2 (Identifier i _ _) = i

isVarPredicate :: Spectre -> Predicate -> Bool
isVarPredicate s (Predicate e1 _ e2) = 
    let cs = proj1 $ findDeclaredClocks s
        vs = proj1 $ findDeclaredVariables s
        e1s = map proj2 $ getAllSymbolsE e1
        e2s = map proj2 $ getAllSymbolsE e2
    in ((intersect cs $ e1s `union` e2s) == []) &&
       ((intersect vs $ e1s `union` e2s) /= [])
    where proj1 :: Maybe [Identifier] -> [String]
          proj1 Nothing = []
          proj1 (Just x) = map proj2 x
          proj2 :: Identifier -> String
          proj2 (Identifier i _ _) = i
          
isValidPredicate :: Spectre -> Predicate -> Bool
isValidPredicate s p = (isClockPredicate s p) || (isVarPredicate s p)         

getInitialUpdates :: Spectre -> [Assignment]
getInitialUpdates s = let (Initially _ (Assignments a)) = getInitially s in a

getInitialLocation :: Spectre -> String
getInitialLocation s = let (Initially (Identifier l _ _)_) = getInitially s in l
